# Test that the atomic corrections of LCAO work correctly,
# by verifying that the different implementations yield the same numbers.
#
# For example the corrections P^* dH P to the Hamiltonian.
#
# This is done by invoking GPAW once for each type of calculation.

from itertools import count

import pytest
from ase.build import molecule, bulk

from gpaw import GPAW, LCAO
from gpaw.mpi import world


def system1():
    system = molecule('CH3CH2OH')
    system.center(vacuum=3.0)
    system.pbc = (0, 1, 1)
    system = system.repeat((1, 1, 2))
    system.rattle(stdev=0.05)
    return system


def system2():
    return bulk('Cu', orthorhombic=True) * (2, 1, 2)


@pytest.mark.later
@pytest.mark.parametrize('atoms, kpts', [
    (system1(), [1, 1, 1]),
    (system2(), [2, 3, 4]),
])
def test_lcao_atomic_corrections(atoms, in_tmp_dir, scalapack, kpts):
    # Use a cell large enough that some overlaps are zero.
    # Thus the matrices will have at least some sparsity.

    corrections = ['dense', 'sparse']

    counter = count()
    energies = []
    for correction in corrections:
        parallel = {}
        if world.size >= 4:
            parallel['band'] = 2
            # if correction.name != 'dense':
            parallel['sl_auto'] = True
        calc = GPAW(mode=LCAO(atomic_correction=correction),
                    basis='sz(dzp)',
                    # spinpol=True,
                    parallel=parallel,
                    txt='gpaw.{}.txt'.format(next(counter)),
                    h=0.35, kpts=kpts,
                    convergence={'maximum iterations': 2})
        atoms.calc = calc
        energy = atoms.get_potential_energy()
        energies.append(energy)
        if calc.world.rank == 0:
            print('e', energy)

    master = calc.wfs.world.rank == 0
    if master:
        print('energies', energies)

    eref = energies[0]
    errs = []
    for energy, c in zip(energies, corrections):
        err = abs(energy - eref)
        errs.append(err)
        if master:
            print('err=%e :: name=%s' % (err, correction))

    maxerr = max(errs)
    assert maxerr < 1e-11, maxerr



import pytest
from gpaw import GPAW, restart
from gpaw.cluster import Cluster
from gpaw.test import equal
from ase.build import molecule
from ase.units import mol, kcal
from gpaw.solvation import SolvationGPAW, get_HW14_water_kwargs


@pytest.fixture
def parameters():
    params = {
        'mode': 'fd',
        'xc': 'PBE',
        'h': 0.24,
        'convergence': {
            'energy': 0.05 / 8.,
            'density': 10.,
            'eigenstates': 10.}}
    return params


@pytest.fixture
def H2O(parameters):
    vac = 4.0

    atoms = Cluster(molecule('H2O'))
    atoms.minimal_box(vac, parameters['h'])

    kwargs = get_HW14_water_kwargs()
    kwargs.update(parameters)
    atoms.calc = SolvationGPAW(**kwargs)
    atoms.get_potential_energy()

    return atoms


def test_solvation_water_water(H2O, parameters):
    SKIP_VAC_CALC = True

    if not SKIP_VAC_CALC:
        atoms = H2O.copy()
        atoms.calc = GPAW(**parameters)
        Evac = atoms.get_potential_energy()
    else:
        # h=0.24, vac=4.0, setups: 0.9.11271, convergence: only energy 0.05 / 8
        Evac = -14.857414548

    Ewater = H2O.get_potential_energy()
    Eelwater = H2O.calc.get_electrostatic_energy()
    Esurfwater = H2O.calc.get_solvation_interaction_energy('surf')
    H2O.get_forces()
    DGSol = (Ewater - Evac) / (kcal / mol)
    print('Delta Gsol: %s kcal / mol' % DGSol)

    equal(DGSol, -6.3, 2.)
    equal(Ewater, Eelwater + Esurfwater, 1e-14)


def test_read(H2O, in_tmp_dir):
    """Read and check some basic properties"""
    fname = 'solvation.gpw'
    H2O.calc.write(fname)
    atoms, calc = restart(fname, Class=SolvationGPAW, txt='-')

    for method in ['get_potential_energy',
                   'get_eigenvalues', 'get_occupation_numbers']:
        assert getattr(calc, method)() == pytest.approx(
            getattr(H2O.calc, method)())

    calc.calculate(atoms)

    for method in ['get_potential_energy',
                   'get_eigenvalues', 'get_occupation_numbers']:
        assert getattr(calc, method)() == pytest.approx(
            getattr(H2O.calc, method)())



"""Test the calculation of the excitation energy of Na2 by RSF and IVOs."""
import pytest
from ase import Atoms
from ase.units import Hartree

import _gpaw
from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.eigensolvers import RMMDIIS
from gpaw.lrtddft import LrTDDFT
from gpaw.mpi import world
from gpaw.occupations import FermiDirac
from gpaw.test import equal, gen


@pytest.mark.hybrids
def test_rsf_yukawa_rsf_ivo_sing_mg(in_tmp_dir, add_cwd_to_setup_paths):
    libxc_version = getattr(_gpaw, 'libxc_version', '2.x.y')
    if int(libxc_version.split('.')[0]) < 3:
        from unittest import SkipTest
        raise SkipTest

    h = 0.35  # Gridspacing
    e_singlet = 4.61
    e_singlet_lr = 5.54

    gen('Mg', xcname='PBE', scalarrel=True, exx=True, yukawa_gamma=0.38)

    c = {'energy': 0.05, 'eigenstates': 3, 'density': 3}
    na2 = Cluster(Atoms('Mg', positions=[[0, 0, 0]]))
    na2.minimal_box(2.5, h=h)
    calc = GPAW(mode='fd',
                txt='mg_ivo.txt',
                xc='LCY-PBE:omega=0.38:excitation=singlet',
                eigensolver=RMMDIIS(), h=h, occupations=FermiDirac(width=0.0),
                spinpol=False, convergence=c)
    na2.calc = calc
    na2.get_potential_energy()
    (eps_homo, eps_lumo) = calc.get_homo_lumo()
    e_ex = eps_lumo - eps_homo
    equal(e_singlet, e_ex, 0.15)
    calc.write('mg.gpw')

    c2 = GPAW('mg.gpw')
    ihomo = int(c2.get_occupation_numbers().sum() / 2 + 0.5) - 1
    assert c2.hamiltonian.xc.excitation == 'singlet'
    lr = LrTDDFT(calc, txt='LCY_TDDFT_Mg.log',
                 restrict={'istart': ihomo, 'jend': ihomo + 1}, nspins=2)
    lr.write('LCY_TDDFT_Mg.ex.gz')
    if world.rank == 0:
        lr2 = LrTDDFT.read('LCY_TDDFT_Mg.ex.gz')
        lr2.diagonalize()
        ex_lr = lr2[1].get_energy() * Hartree
        equal(e_singlet_lr, ex_lr, 0.15)



from ase.constraints import FixBondLengths
from ase.calculators.tip3p import TIP3P
from ase.calculators.tip3p import qH, sigma0, epsilon0

from _gpaw import adjust_positions, adjust_momenta, calculate_forces_H2O
from ase.calculators.calculator import Calculator, all_changes

import numpy as np

A = 4 * epsilon0 * sigma0**12
B = -4 * epsilon0 * sigma0**6


class TIP3PWaterModel(TIP3P):
    def calculate(self, atoms=None,
                  properties=['energy'],
                  system_changes=all_changes):

        Calculator.calculate(self, atoms, properties, system_changes)

        R = self.atoms.positions.reshape((-1, 3, 3))
        Z = self.atoms.numbers
        pbc = self.atoms.pbc
        diagcell = self.atoms.cell.diagonal()
        nh2o = len(R)

        assert (self.atoms.cell == np.diag(diagcell)).all(), 'not orthorhombic'
        assert ((diagcell >= 2 * self.rc) | ~pbc).all(), 'cutoff too large'
        if Z[0] == 8:
            o = 0
        else:
            o = 2

        assert o == 0
        assert (Z[o::3] == 8).all()
        assert (Z[(o + 1) % 3::3] == 1).all()
        assert (Z[(o + 2) % 3::3] == 1).all()

        charges = np.array([qH, qH, qH])
        charges[o] *= -2

        energy = 0.0
        forces = np.zeros((3 * nh2o, 3))

        cellmat = np.zeros((3, 3))
        np.fill_diagonal(cellmat, diagcell)   # c code wants 3x3 cell ...

        energy += calculate_forces_H2O(np.array(atoms.pbc, dtype=np.uint8),
                                       cellmat, A, B, self.rc, self.width,
                                       charges, self.atoms.get_positions(),
                                       forces)

        if self.pcpot:
            e, f = self.pcpot.calculate(np.tile(charges, nh2o),
                                        self.atoms.positions)
            energy += e
            forces += f

        self.results['energy'] = energy
        self.results['forces'] = forces


class FixBondLengthsWaterModel(FixBondLengths):

    def __init__(self, pairs, tolerance=1e-13, bondlengths=None,
                 iterations=None):
        FixBondLengths.__init__(self, pairs, tolerance=tolerance,
                                bondlengths=bondlengths,
                                iterations=iterations)

    def initialize_bond_lengths(self, atoms):
        bondlengths = FixBondLengths.initialize_bond_lengths(self, atoms)
        # Make sure that the constraints are compatible with the C-code
        assert len(self.pairs) % 3 == 0
        masses = atoms.get_masses()

        self.start = self.pairs[0][0]
        self.end = self.pairs[-3][0]
        self.NW = (self.end - self.start) // 3 + 1
        assert (self.end - self.start) % 3 == 0
        for i in range(self.NW):
            for j in range(3):
                assert self.pairs[i * 3 + j][0] == self.start + i * 3 + j
                assert self.pairs[i * 3 + j][1] == (self.start +
                                                    i * 3 + (j + 1) % 3)
                assert abs(bondlengths[i * 3 + j] - bondlengths[j]) < 1e-6
                assert masses[i * 3 + j + self.start] == masses[j + self.start]
        return bondlengths

    def select_indices(self, r):
        return r[self.start:self.start + self.NW * 3, :]

    def adjust_positions(self, atoms, new):
        masses = atoms.get_masses()

        if self.bondlengths is None:
            self.bondlengths = self.initialize_bond_lengths(atoms)

        return adjust_positions(self.bondlengths[:3],
                                masses[self.start:self.start + 3],
                                self.select_indices(atoms.get_positions()),
                                self.select_indices(new))

    def adjust_momenta(self, atoms, p):
        masses = atoms.get_masses()

        if self.bondlengths is None:
            self.bondlengths = self.initialize_bond_lengths(atoms)

        return adjust_momenta(masses[self.start:self.start + 3],
                              self.select_indices(atoms.get_positions()),
                              self.select_indices(p))

    def index_shuffle(self, atoms, ind):
        raise NotImplementedError



# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

from math import sqrt, cos, sin

import numpy as np

from gpaw.spherical_harmonics import Y


s = sqrt(0.5)
t = sqrt(3) / 3
# Points on the unit sphere:
sphere_lm = [
    np.array([(1, 0, 0)]),  # s
    np.array([(1, 0, 0), (0, 1, 0), (0, 0, 1)]),  # p
    np.array([(s, s, 0), (0, s, s), (s, 0, s), (1, 0, 0), (0, 0, 1)]),  # d
    np.array([(s, s, 0), (0, s, s), (s, 0, s),
              (1, 0, 0), (s, -s, 0),
              (t, -t, t), (t, t, -t)])]  # f


def Y_matrix(l, U_vv):
    """YMatrix(l, symmetry) -> matrix.

    For 2*l+1 points on the unit sphere (m1=0,...,2*l) calculate the
    value of Y_lm2 for m2=0,...,2*l.  The points are those from the
    list sphere_lm[l] rotated as described by U_vv."""

    Y_mm = np.zeros((2 * l + 1, 2 * l + 1))
    for m1, point in enumerate(sphere_lm[l]):
        x, y, z = np.dot(point, U_vv)
        for m2 in range(2 * l + 1):
            L = l**2 + m2
            Y_mm[m1, m2] = Y(L, x, y, z)
    return Y_mm


identity = np.eye(3)
iY_lmm = [np.linalg.inv(Y_matrix(l, identity)) for l in range(4)]


def rotation(l, U_vv):
    """Rotation(l, symmetry) -> transformation matrix.

    Find the transformation from Y_lm1 to Y_lm2."""

    return np.dot(iY_lmm[l], Y_matrix(l, U_vv))


def Y_rotation(l, angle):
    Y_mm = np.zeros((2 * l + 1, 2 * l + 1))
    sn = sin(angle)
    cs = cos(angle)
    for m1, point in enumerate(sphere_lm[l]):
        x = point[0]
        y = cs * point[1] - sn * point[2]
        z = sn * point[1] + cs * point[2]
        for m2 in range(2 * l + 1):
            L = l**2 + m2
            Y_mm[m1, m2] = Y(L, x, y, z)
    return Y_mm


del s, identity



import pytest
import numpy as np
from gpaw.response.coulomb_kernels import CoulombKernel
from gpaw.response.pair import get_gs_and_context
from gpaw import GPAW, PW, FermiDirac
from gpaw.mpi import world
from gpaw.response.chi0 import Chi0
from gpaw.response.screened_interaction import initialize_w_calculator
from ase.build import bulk
from ase import Atoms


@pytest.mark.response
def test_Wsymm(in_tmp_dir, scalapack):
    def calc_gs(symm):
        k = 2
        cell = bulk('Ga', 'fcc', a=5.68).cell
        a = Atoms('GaAs', cell=cell, pbc=True,
                  scaled_positions=((0, 0, 0), (0.25, 0.25, 0.25)))
        
        # First with symmetry off
        calc = GPAW(mode=PW(400),
                    xc='LDA',
                    occupations=FermiDirac(width=0.01),
                    convergence={'bands': -4},
                    symmetry=symm,
                    kpts={'size': (k, k, k), 'gamma': True},
                    txt='gs_GaAs.txt')

        a.calc = calc
        a.get_potential_energy()
        if symm == 'off':
            calc.write('GaAs.gpw', mode='all')
        else:
            calc.write('GaAs_symm.gpw', mode='all')
        return calc
    
    def get_IBZ_k(calc):
        gs = calc.gs_adapter()
        qclist = gs.kd.ibzk_kc
        return qclist

    def calc_W(seed, q_c_list):
        omega = np.array([0])
        chi0calc = Chi0(seed + '.gpw',
                        frequencies=omega,
                        hilbert=False,
                        ecut=100,
                        txt='test.log',
                        intraband=False)
        txt = 'out.txt'
        gs, wcontext = get_gs_and_context(
            seed + '.gpw',
            txt, world=world,
            timer=None)
        truncation = None
        coulomb = CoulombKernel.from_gs(gs, truncation=truncation)
        wcalc = initialize_w_calculator(chi0calc,
                                        wcontext,
                                        coulomb=coulomb,
                                        integrate_gamma=0)
        Wlist = []
        qlist = []
        for iq, q_c in enumerate(q_c_list):
            chi0 = chi0calc.calculate(q_c)
            W_wGG = wcalc.calculate_W_wGG(chi0,
                                          fxc_mode='GW',
                                          only_correlation=False)
            Wlist.append(W_wGG)
            qlist.append(q_c)
        return Wlist, qlist

    for symm in ['off', {}]:
        # calc gs with and without symmetry
        calc = calc_gs(symm)
        # get list of Q in IBZ with symmetry
        q_c_list = get_IBZ_k(calc)

    # calc W_GG
    WGG_nosymm, qnosymm = calc_W('GaAs', q_c_list)
    WGG_symm, qsymm = calc_W('GaAs_symm', q_c_list)

    # compare W_GG for all k in IBZ
    for iq, q_c in enumerate(q_c_list):
        assert np.allclose(qnosymm[iq], qsymm[iq])
        assert np.allclose(WGG_symm[iq], WGG_nosymm[iq])



from itertools import product

import pytest
import numpy as np

from gpaw.response.integrators import (TetrahedronIntegrator, Integrand,
                                       HilbertTetrahedron)
from gpaw.response.frequencies import FrequencyGridDescriptor

from gpaw.response import ResponseContext


class MyIntegrand(Integrand):
    def matrix_element(self, x_c):
        return np.array([[1.]], complex)

    def eigenvalues(self, x_c):
        return np.array([(x_c**2).sum()**0.5], float)


@pytest.mark.response
def test_tetrahedron_integrator():
    cell_cv = np.eye(3)
    context = ResponseContext()
    integrator = TetrahedronIntegrator(cell_cv, context, nblocks=1)
    x_g = np.linspace(-1, 1, 30)
    x_gc = np.array([comb for comb in product(*([x_g] * 3))])

    domain = (x_gc,)
    out_wxx = np.zeros((1, 1, 1), complex)
    integrator.integrate(task=HilbertTetrahedron(integrator.blockcomm),
                         domain=domain,
                         integrand=MyIntegrand(),
                         wd=FrequencyGridDescriptor([-1.0]),
                         out_wxx=out_wxx)

    assert abs(out_wxx[0, 0, 0] - 4 * np.pi) < 1e-2
    # equal(out_wxx[0, 0, 0], 4 * np.pi, 1e-2,
    #       msg='Integrated area of unit sphere is not 4 * pi')



import numpy as np
from gpaw.xc.kernel import XCKernel


class LB94(XCKernel):
    """Correction to LDA to resemble asymptotic -1/r potential.

    See:

        van Leeuwen and Baerends, Phys.Rev.A vol 49 (1994) 2421
    """

    def __init__(self, beta=0.05):
        XCKernel.__init__(self, 'LDA')
        self.name = 'LB94'
        self.type = 'GGA'
        self.beta = beta

    def calculate(self, e_g, n_sg, dedn_sg,
                  sigma_xg=None, dedsigma_xg=None,
                  tau_sg=None, dedtau_sg=None):
        XCKernel.calculate(self, e_g, n_sg, dedn_sg, sigma_xg, dedsigma_xg)
        for s, n_g in enumerate(n_sg):
            n_g = n_g * len(n_sg)
            n_g[n_g < 1e-10] = 1e-10
            y_g = n_g**(1 / 3.0)
            x_g = sigma_xg[2 * s]**0.5 / (y_g * n_g) * len(n_sg)
            x_g[x_g > 500] = 0.0
            dedn_sg[s] -= (self.beta * x_g**2 * y_g /
                           (1 + 3 * self.beta * x_g * np.arcsinh(x_g)))
        dedsigma_xg[:] = 0.0



"""Unified pseudopotential format (UPF).

UPF refers to a collection of different formats to store
pseudopotentials or PAW setups.  This file attempts to parse them and
provide pseudopotential objects for use with GPAW.
"""

from optparse import OptionParser
from xml.etree.ElementTree import parse as xmlparse, fromstring
from xml.etree.ElementTree import ParseError

import numpy as np
from ase.data import atomic_numbers

from gpaw.atom.atompaw import AtomPAW
from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
from gpaw.setup_data import search_for_file
from gpaw.basis_data import Basis, BasisFunction
from gpaw.pseudopotential import (PseudoPotential, screen_potential,
                                  figure_out_valence_states,
                                  get_radial_hartree_energy)
from gpaw.spline import Spline
from gpaw.utilities import pack2, divrl


class UPFStateSpec:
    def __init__(self, index, label, l, values, occupation=None, n=None):
        self.index = index
        self.label = label
        self.l = l
        self.values = values
        self.occupation = occupation
        self.n = n


class UPFCompensationChargeSpec:  # Not used right now.......
    def __init__(self, i, j, l, qint, values, coefs):
        self.i = i
        self.j = j
        self.l = l
        self.qint = qint
        self.values = values
        self.coefs = coefs

    def __repr__(self):
        return ('CompensationCharge(i=%d, j=%d, l=%d, qint=%s, '
                'ccvalues=[...], coefs=[...(%d)...]') % (self.i, self.j,
                                                         self.l, self.qint,
                                                         len(self.coefs))


def trim_outer_zeros(values, threshold=0.0):
    values = list(values)
    if abs(values[-1]) <= threshold:
        while abs(values[-2]) <= threshold:
            values.pop()
        assert abs(values[-1]) <= threshold
    # This will give an error if the array length becomes smaller than 2.
    # Which would be fine I guess.
    return np.array(values)


def parse_upf(fname):
    """Parse UPF pseudopotential from file descriptor.

    Return a dictionary with the parsed data.

    UPF is a custom format with some XML tags in it.  The files are
    not well-formed XML.  We will just do some custom parsing rather
    than involving an xml library.
    """

    pp = {}

    try:
        root = xmlparse(fname).getroot()
    except ParseError:
        # Typical!  Non-well-formed file full of probable FORTRAN output.
        # We'll try to insert our own header and see if things go well.
        root = fromstring('\n'.join(['<xml>', open(fname).read(), '</xml>']))

    pp['fname'] = fname

    header_element = root.find('PP_HEADER')
    header = header_element.attrib

    # v201 is the sensible version.
    # There are other files out there and we will have to deal with their
    # inconsistencies in the most horrendous way conceivable: by if-statements.
    v201 = (root.attrib.get('version') == '2.0.1')

    def toarray(element):
        attr = element.attrib
        numbers = [float(n) for n in element.text.split()]
        if attr:
            assert attr['type'] == 'real'
            assert int(attr['size']) == len(numbers)
            icut = attr.get('cutoff_radius_index')
            # XXX There's also something called ultrasoft_cutoff_radius
            if icut is not None:
                numbers = numbers[:int(icut)]

        return np.array(numbers)

    if v201:
        for key, val in header.items():
            header[key] = val.strip()  # some values have whitespace...
        for key in ['is_paw', 'is_coulomb', 'has_so', 'has_wfc', 'has_gipaw',
                    'paw_as_gipaw', 'core_correction']:
            if key in header:
                header[key] = {'F': False, 'T': True}[header[key]]
        for key in ['z_valence', 'total_psenergy', 'wfc_cutoff', 'rho_cutoff']:
            if key not in header:
                continue
            header[key] = float(header[key])
        for key in ['l_max', 'l_max_rho', 'mesh_size', 'number_of_wfc',
                    'number_of_proj']:
            if key not in header:
                continue
            header[key] = int(header[key])
    else:
        assert len(header) == 0
        # This is the crappy format.
        headerlines = [line for line
                       in header_element.text.strip().splitlines()]
        # header['version'] = headerlines[0].split()[0]
        header['element'] = headerlines[1].split()[0]
        # header['has_nlcc'] = {'T': True, 'F': False} .........
        # assert not header['has_nlcc']
        # XC functional?
        header['z_valence'] = float(headerlines[5].split()[0])
        header['total_psenergy'] = float(headerlines[6].split()[0])
        # cutoffs?
        header['l_max'] = int(headerlines[8].split()[0])
        header['mesh_size'] = int(headerlines[9].split()[0])
        nwfs, nprojectors = headerlines[10].split()[:2]
        header['number_of_wfc'] = int(nwfs)
        header['number_of_proj'] = int(nprojectors)

    info_element = root.find('PP_INFO')
    pp['info'] = info_element.text

    pp['header'] = header
    mesh_element = root.find('PP_MESH')
    pp['r'] = toarray(mesh_element.find('PP_R'))
    pp['rab'] = toarray(mesh_element.find('PP_RAB'))

    # Convert to Hartree from Rydberg.
    pp['vlocal'] = 0.5 * toarray(root.find('PP_LOCAL'))

    non_local = root.find('PP_NONLOCAL')

    pp['projectors'] = []
    for element in non_local:
        if element.tag.startswith('PP_BETA'):
            if '.' in element.tag:
                name, num = element.tag.split('.')
                assert name == 'PP_BETA'
                attr = element.attrib
                proj = UPFStateSpec(int(attr['index']),
                                    attr.get('label'),
                                    int(attr['angular_momentum']),
                                    toarray(element))
                assert num == attr['index']
            else:
                tokens = element.text.split()
                metadata = tokens[:5]
                values = map(float, tokens[5:])

                npts = int(metadata[4])
                assert npts == len(values), (npts, len(values))
                while values[-1] == 0.0:
                    values.pop()

                proj = UPFStateSpec(int(tokens[0]), '??', int(tokens[1]),
                                    np.array(values))

            pp['projectors'].append(proj)
        else:
            if v201:
                assert element.tag == 'PP_DIJ', element.tag
                # XXX probably measured in Rydberg.
                pp['DIJ'] = 0.5 * toarray(element)
            else:
                lines = element.text.strip().splitlines()
                nnonzero = int(lines[0].split()[0])
                nproj = pp['header']['number_of_proj']
                D_ij = np.zeros((nproj, nproj))
                for n in range(1, nnonzero + 1):
                    tokens = lines[n].split()
                    i = int(tokens[0]) - 1
                    j = int(tokens[1]) - 1
                    D = float(tokens[2])
                    D_ij[i, j] = D
                    D_ij[j, i] = D
                assert len(lines) == 1 + nnonzero
                # XXX probably measured in Rydberg.
                pp['DIJ'] = 0.5 * D_ij

    pswfc_element = root.find('PP_PSWFC')
    pp['states'] = []
    if v201:
        for element in pswfc_element:
            attr = element.attrib
            name = element.tag
            state = UPFStateSpec(int(attr['index']),
                                 attr['label'],
                                 int(attr['l']),
                                 toarray(element),
                                 float(attr['occupation']),
                                 int(attr['n']))
            pp['states'].append(state)
    else:
        state_data = []
        for line in pswfc_element.text.splitlines()[1:]:
            if line.endswith('Wavefunction'):
                values = []
                state_data.append((line, values))
            else:
                values.extend(map(float, line.split()))
        for header, values in state_data:
            label, l, occupation, wfstring = header.split()
            assert wfstring == 'Wavefunction'
            state = UPFStateSpec(None, label, int(l), np.array(values),
                                 occupation=float(occupation))
            pp['states'].append(state)
        # print repr(lines[0])
    # assert len(pp['states']) > 0

    pp['rhoatom'] = toarray(root.find('PP_RHOATOM'))
    return pp


sg15_special_valence_states = {
    'Re': ([5, 5, 6, 5], [0, 1, 0, 2], [2., 6., 2., 5.]),
    'Os': ([5, 5, 6, 5], [0, 1, 0, 2], [2., 6., 2., 6.]),
    'Ir': ([5, 5, 6, 5], [0, 1, 0, 2], [2., 6., 2., 7.])}


def read_sg15(fname):
    data = parse_upf(fname)
    symbol = data['header']['element']
    valence_states = None

    states_nlf = sg15_special_valence_states.get(symbol)
    if states_nlf is not None:
        valence_states = [UPFStateSpec(index=None, label=None, n=n, l=l,
                                       values=None, occupation=f)
                          for n, l, f in zip(*states_nlf)]

    data = UPFSetupData(data, filename=fname, valence_states=valence_states)
    return data


class UPFSetupData:
    def __init__(self, data, valence_states=None, filename=None):
        # data can be string (filename)
        # or dict (that's what we are looking for).
        # Maybe just a symbol would also be fine if we know the
        # filename to look for.
        if isinstance(data, str):
            filename = data
            data = parse_upf(data)
        elif filename is None:
            filename = '[N/A]'

        self.filename = filename

        assert isinstance(data, dict)
        self.data = data  # more or less "raw" data from the file

        self.name = 'upf'

        header = data['header']

        self.symbol = header['element']
        self.Z = atomic_numbers[self.symbol]
        self.Nv = header['z_valence']
        self.Nc = self.Z - self.Nv
        self.Delta0 = -self.Nv / np.sqrt(4.0 * np.pi)  # like hgh
        self.rcut_j = [data['r'][len(proj.values) - 1]
                       for proj in data['projectors']]

        # beta = 0.1 # XXX nice parameters?
        # N = 4 * 450 # XXX
        # This is "stolen" from hgh.  Figure out something reasonable
        # rgd = AERadialGridDescriptor(beta / N, 1.0 / N, N,
        #                              default_spline_points=100)

        from gpaw.atom.radialgd import EquidistantRadialGridDescriptor
        rgd = EquidistantRadialGridDescriptor(0.02)
        self.rgd = rgd

        # Whyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy???
        # What abominable part of the code requires the states
        # to be ordered like this?
        self.jargs = self.get_jargs()

        projectors = [data['projectors'][jarg] for jarg in self.jargs]
        if projectors:
            self.l_j = [proj.l for proj in projectors]
            self.pt_jg = []
            for proj in projectors:
                val = proj.values.copy()
                val /= 2.
                val[1:] /= data['r'][1:len(val)]
                val[0] = val[1]
                pt_g = self._interp(val)  # * np.sqrt(4.0 * np.pi)
                # sqrnorm = (pt_g**2 * self.rgd.dr_g).sum()
                self.pt_jg.append(pt_g)

        else:
            self.l_j = [0]
            gcut = self.rgd.r2g(1.0)
            self.pt_jg = [np.zeros(gcut)]  # XXX yet another "null" function

        self.rcgauss = 0.0  # XXX .... what is this used for?
        self.ni = sum([2 * l + 1 for l in self.l_j])

        self.fingerprint = None  # XXX hexdigest the file?
        self.lq = None  # XXX

        if valence_states is None:
            valence_states = data['states']

        if valence_states:
            states_lmax = max([state.l for state in valence_states])
            f_ln = [[] for _ in range(1 + states_lmax)]
            electroncount = 0.0
            for state in valence_states:
                # Where should the electrons be in the inner list??
                # This is probably wrong and will lead to bad initialization
                f_ln[state.l].append(state.occupation)
                electroncount += state.occupation
                # The Cl.pz-hgh.UPF from quantum espresso has only 6
                # but should have 7 electrons.  Oh well....
            # err = abs(electroncount - self.Nv)
            self.f_j = [state.occupation for state in valence_states]
            self.n_j = [state.n for state in valence_states]
            self.l_orb_J = [state.l for state in valence_states]
            self.f_ln = f_ln
        else:
            self.n_j, self.l_orb_J, self.f_j, self.f_ln = \
                figure_out_valence_states(self)

        vlocal_unscreened = data['vlocal']

        # The UPF representation of HGH setups should be equal to that
        # used with setups='hgh'.  But the UPF files do not contain
        # info on the localization of the compensation charges!  That
        # means the screen_potential code will choose the shape of the
        # compensation charges, resulting in some numerical differences.
        #
        # One could hack that it looks up some radii to compare e.g. H2O.
        # The results of the below comments still have some numerical error.
        # But it's not huge.
        # a = None
        # if self.symbol == 'H':
        #     a = 0.2
        # elif self.symbol == 'O':
        #     a = .247621

        vbar_g, ghat_g = screen_potential(data['r'], vlocal_unscreened,
                                          self.Nv)
        self.Eh_compcharge = get_radial_hartree_energy(data['r'][:len(ghat_g)],
                                                       ghat_g)
        self.vbar_g = self._interp(vbar_g) * np.sqrt(4.0 * np.pi)
        self.ghat_lg = [4.0 * np.pi / self.Nv * self._interp(ghat_g)]

    def get_jargs(self):
        projectors = list(self.data['projectors'])
        jargs = []

        for n in range(4):
            for l in range(4):
                for i, proj in enumerate(projectors):
                    if proj.l == l:
                        jargs.append(proj.index - 1)
                        projectors.pop(i)
                        break
        return jargs

    def tostring(self):
        lines = []
        indent = 0

        def add(line):
            lines.append(indent * ' ' + line)

        add('Norm-conserving UPF setup:')
        indent = 2
        add('Element: %4s' % self.symbol)
        add('Z:       %4s' % self.Z)
        add('Valence: %4s' % self.Nv)
        indent -= 2
        add('Projectors:')
        indent += 2
        for j, proj in enumerate(self.data['projectors']):
            add('l=%d rcut=%s' % (proj.l, self.rcut_j[j]))
        indent -= 2
        if len(self.data['states']) == 0:
            add('No states stored on this setup')
        else:
            add('States:')
            indent += 2
            for j, state in enumerate(self.data['states']):
                add('l=%d f=%s' % (state.l, state.occupation))
            indent -= 2
        add('Local potential cutoff: %s'
            % self.rgd.r_g[len(self.vbar_g) - 1])
        add('Comp charge cutoff:     %s'
            % self.rgd.r_g[len(self.ghat_lg[0]) - 1])
        add('File: %s' % self.filename)
        add('')
        return '\n'.join(lines)

    def print_info(self, txt, setup):
        txt(self.tostring())

    # XXX something can perhaps be stolen from HGH
    def expand_hamiltonian_matrix(self):
        """Construct K_p from individual h_nn for each l."""
        ni = sum([2 * l + 1 for l in self.l_j])
        nj = len(self.l_j)

        H_ii = np.zeros((ni, ni))
        if len(self.data['DIJ']) == 0:
            return pack2(H_ii)

        # Multiply by 4.
        # I think the factor of 4 compensates for the fact that the projectors
        # all had square norms of 4, but we brought them back down to 1
        # because that's more sensible.
        H_jj = 4.0 * self.data['DIJ'].reshape((nj, nj))
        m1start = 0
        for j1, l1 in enumerate(self.l_j):
            j1 = self.jargs[j1]
            m1stop = m1start + 2 * l1 + 1
            m2start = 0
            for j2, l2 in enumerate(self.l_j):
                j2 = self.jargs[j2]
                m2stop = m2start + 2 * l2 + 1
                if l1 == l2:
                    dm = m1stop - m1start
                    H_ii[m1start:m1stop, m2start:m2stop] = \
                        np.eye(dm) * H_jj[j1, j2]
                else:
                    assert H_jj[j1, j2] == 0.0
                m2start = m2stop
            m1start = m1stop
        return pack2(H_ii)

    def get_local_potential(self):
        vbar = Spline(0, self.rgd.r_g[len(self.vbar_g) - 1], self.vbar_g)
        return vbar

    # XXXXXXXXXXXXXXXXX stolen from hghsetupdataf
    def get_projectors(self):
        # XXX equal-range projectors still required for some reason
        maxlen = max([len(pt_g) for pt_g in self.pt_jg])
        pt_j = []
        for l, pt1_g in zip(self.l_j, self.pt_jg):
            pt2_g = self.rgd.zeros()[:maxlen]
            pt2_g[:len(pt1_g)] = divrl(pt1_g, l, self.rgd.r_g[:len(pt1_g)])
            spline = Spline(l, self.rgd.r_g[maxlen - 1], pt2_g)
            pt_j.append(spline)
        return pt_j

    def _interp(self, func):
        r_g = self.rgd.r_g
        gcut = len(func)
        rcut = self.data['r'][gcut - 1]
        gcutnew = np.searchsorted(self.rgd.r_g, rcut)
        rnew = r_g[:gcutnew]
        ynew = np.interp(rnew, self.data['r'][:gcut], func, right=0.0)
        return ynew

    def get_compensation_charge_functions(self):
        assert len(self.ghat_lg) == 1
        ghat_g = self.ghat_lg[0]
        ng = len(ghat_g)
        rcutcc = self.rgd.r_g[ng - 1]  # correct or not?
        r = np.linspace(0.0, rcutcc, 50)
        ghat_g[-1] = 0.0
        ghatnew_g = Spline(0, rcutcc, ghat_g).map(r)
        return r, [0], [ghatnew_g]

    def create_basis_functions(self):
        if len(self.data['states']) > 0:
            return self.get_stored_basis_functions()
        else:
            from gpaw.pseudopotential import generate_basis_functions
            return generate_basis_functions(self)

    def get_stored_basis_functions(self, ):
        states = self.data['states']
        maxlen = max([len(state.values) for state in states])
        orig_r = self.data['r']
        rcut = min(orig_r[maxlen - 1], 12.0)  # XXX hardcoded 12 max radius

        d = 0.02
        ng = int(1 + rcut / d)
        rgd = EquidistantRadialGridDescriptor(d, ng)

        b = Basis(self.symbol, 'upf', readxml=False, rgd=rgd)
        b.generatordata = 'upf-pregenerated'

        for j, state in enumerate(states):
            val = state.values
            phit_g = np.interp(rgd.r_g, orig_r, val)
            phit_g = divrl(phit_g, 1, rgd.r_g)
            icut = len(phit_g) - 1  # XXX correct or off-by-one?
            rcut = rgd.r_g[icut]
            bf = BasisFunction(None, state.l, rcut, phit_g, 'pregenerated')
            b.bf_j.append(bf)
        return b

    def build(self, xcfunc, lmax, basis, filter=None):
        # XXX better to create basis functions after filtering?
        # Although basis functions are not meant for same grid
        if basis is None:
            basis = self.create_basis_functions()
        return PseudoPotential(self, basis, filter)


def main_plot():
    p = OptionParser(usage='%prog [OPTION] [FILE...]',
                     description='plot upf pseudopotential from file.')
    p.add_option('--calculate', action='store_true',
                 help='calculate density and orbitals.')
    opts, args = p.parse_args()

    import matplotlib.pyplot as plt
    for arg in args:
        if not arg.lower().endswith('.upf'):
            # This is a bit bug prone.  It is subject to files
            # "sneaking in" unexpectedly due to '*' from
            # higher-priority search directories.  Then again, this
            # probably will not happen, and the runtime setup loading
            # mechanism is the same anyway so at least it is honest.
            fname, source = search_for_file('%s*.[uU][pP][fF]' % arg)
            if fname is None:
                p.error('No match within search paths: %s' % arg)
        else:
            fname = arg
        pp = parse_upf(fname)
        print('--- %s ---' % fname)
        print(UPFSetupData(pp).tostring())
        print(pp['info'])
        upfplot(pp, show=False, calculate=opts.calculate)
    plt.show()


def upfplot(setup, show=True, calculate=False):
    # A version of this, perhaps nicer, is in pseudopotential.py.
    # Maybe it is not worth keeping this version
    if isinstance(setup, dict):
        setup = UPFSetupData(setup)

    pp = setup.data
    r0 = pp['r'].copy()
    r0[0] = 1e-8

    def rtrunc(array, rdividepower=0):
        r = r0[:len(array)]
        arr = divrl(array, rdividepower, r)
        return r, arr

    import matplotlib.pyplot as plt
    fig = plt.figure()
    fig.canvas.set_window_title('%s - UPF setup for %s' % (pp['fname'],
                                                           setup.symbol))

    vax = fig.add_subplot(221)
    pax = fig.add_subplot(222)
    rhoax = fig.add_subplot(223)
    wfsax = fig.add_subplot(224)

    r, v = rtrunc(pp['vlocal'])

    vax.plot(r, v, label='vloc')

    vscreened, rhocomp = screen_potential(r, v, setup.Nv)
    icut = len(rhocomp)
    vcomp = v.copy()
    vcomp[:icut] -= vscreened
    vax.axvline(r[icut], ls=':', color='k')

    vax.plot(r, vcomp, label='vcomp')
    vax.plot(r[:icut], vscreened, label='vscr')
    vax.axis(xmin=0.0, xmax=6.0)
    rhoax.plot(r[:icut], rhocomp[:icut], label='rhocomp')

    for j, proj in enumerate(pp['projectors']):
        r, p = rtrunc(proj.values, 0)
        pax.plot(r, p,
                 label='p%d [l=%d]' % (j + 1, proj.l))

    for j, st in enumerate(pp['states']):
        r, psi = rtrunc(st.values, 1)
        wfsax.plot(r, r * psi, label='wf%d %s' % (j, st.label))

    r, rho = rtrunc(pp['rhoatom'], 2)
    wfsax.plot(r, r * rho, label='rho')

    if calculate:
        calc = AtomPAW(setup.symbol,
                       [setup.f_ln],
                       # xc='PBE', # XXX does not support GGAs :( :( :(
                       setups={setup.symbol: setup},
                       h=0.08,
                       rcut=10.0)
        r_g = calc.wfs.gd.r_g
        basis = calc.extract_basis_functions()
        wfsax.plot(r_g, r_g * calc.density.nt_sg[0] * 4.0 * np.pi,
                   ls='--', label='rho [calc]', color='r')

        splines = basis.tosplines()
        for spline, bf in zip(splines, basis.bf_j):
            wfsax.plot(r_g, r_g * spline.map(r_g), label=bf.type)

    vax.legend(loc='best')
    rhoax.legend(loc='best')
    pax.legend(loc='best')
    wfsax.legend(loc='best')

    for ax in [vax, rhoax, pax, wfsax]:
        ax.set_xlabel('r [Bohr]')
        ax.axhline(0.0, ls=':', color='black')

    vax.set_ylabel('potential')
    pax.set_ylabel('projectors')
    wfsax.set_ylabel(r'$r \psi(r), r n(r)$')
    rhoax.set_ylabel('Comp charges')

    fig.subplots_adjust(wspace=0.3)

    if show:
        plt.show()



import pytest
from ase import Atom, Atoms
from ase.units import Bohr
from ase.parallel import parprint
from gpaw import GPAW
from gpaw.test import equal
from gpaw.lrtddft import LrTDDFT
from gpaw.mpi import world
from gpaw.lrtddft.excited_state import ExcitedState


@pytest.fixture
def H2struct():
    R = 0.7  # approx. experimental bond length
    a = 3.0
    c = 4.0
    return Atoms([Atom('H', (a / 2, a / 2, (c - R) / 2)),
                  Atom('H', (a / 2, a / 2, (c + R) / 2))],
                 cell=(a, a, c))


@pytest.fixture
def H2(H2struct):
    H2 = H2struct.copy()
    H2.calc = GPAW(mode='fd', xc='PBE',
                   poissonsolver={'name': 'fd'},
                   nbands=3, spinpol=False)
    H2.get_potential_energy()
    return H2


@pytest.fixture
def H2spin(H2struct):
    H2 = H2struct.copy()
    H2.calc = GPAW(mode='fd', xc='PBE', nbands=2,
                   poissonsolver={'name': 'fd'},
                   spinpol=True, parallel={'domain': world.size})
    H2.get_potential_energy()
    return H2


@pytest.fixture
def lr(H2):
    """LrTDDFT object without spin"""
    lr = LrTDDFT(H2.calc, xc='LDA')
    lr.diagonalize()
    return lr


@pytest.fixture
def lr_vspin(H2):
    lr = LrTDDFT(H2.calc, xc='LDA', nspins=2)
    lr.diagonalize()
    return lr


@pytest.fixture
def lr_spin(H2spin):
    lr = LrTDDFT(H2spin.calc, xc='LDA', nspins=2)
    lr.diagonalize()
    return lr


@pytest.mark.lrtddft
def test_finegrid(H2, lr):
    for finegrid in [1, 0]:
        lr2 = LrTDDFT(H2.calc, xc='LDA', finegrid=finegrid)
        lr2.diagonalize()
        parprint('finegrid, t1, t3=', finegrid, lr[0], lr2[0])
        equal(lr[0].get_energy(), lr2[0].get_energy(), 5.e-4)


@pytest.mark.lrtddft
def test_velocity_form(lr):
    for ozr, ozv in zip(lr[0].get_oscillator_strength(),
                        lr[0].get_oscillator_strength('v')):
        equal(ozr, ozv, 0.1)


@pytest.mark.lrtddft
def test_singlet_triplet(lr_vspin, lr_spin):
    # singlet/triplet separation
    singlet, triplet = lr_vspin.singlets_triplets()

    # singlet/triplet separation
    precision = 1.e-5
    singlet.diagonalize()
    equal(singlet[0].get_energy(), lr_spin[1].get_energy(), precision)
    equal(singlet[0].get_oscillator_strength()[0],
          lr_spin[1].get_oscillator_strength()[0], precision)
    triplet.diagonalize()
    equal(triplet[0].get_oscillator_strength()[0], 0)
    equal(triplet[0].get_energy(), lr_spin[0].get_energy(), precision)
    equal(triplet[0].get_oscillator_strength()[0], 0)


@pytest.mark.lrtddft
def test_spin(lr, lr_vspin, lr_spin):
    # without spin
    t1 = lr[0]
    ex = ExcitedState(lr, 0)
    den = ex.get_pseudo_density() * Bohr**3

    # with spin

    # the triplet is lower, so that the second is the first singlet
    # excited state
    t2 = lr_vspin[1]
    ex_vspin = ExcitedState(lr_vspin, 1)
    den_vspin = ex_vspin.get_pseudo_density() * Bohr**3

    parprint('with virtual/wo spin t2, t1=',
             t2.get_energy(), t1 .get_energy())
    equal(t1.get_energy(), t2.get_energy(), 5.e-7)
    gd = lr.calculator.density.gd
    finegd = lr.calculator.density.finegd
    ddiff = gd.integrate(abs(den - den_vspin))
    parprint('   density integral, difference=',
             gd.integrate(den), gd.integrate(den_vspin), ddiff)
    parprint('   aed density integral',
             finegd.integrate(ex.get_all_electron_density() * Bohr**3),
             finegd.integrate(ex_vspin.get_all_electron_density() *
                              Bohr**3))
    assert ddiff < 1.e-4

    for i in range(2):
        parprint('i, real, virtual spin: ', i, lr_vspin[i], lr_spin[i])
        equal(lr_vspin[i].get_energy(), lr_spin[i].get_energy(), 6.e-6)
        ex_vspin = ExcitedState(lr_vspin, i)
        den_vspin = ex_vspin.get_pseudo_density() * Bohr**3
        ex_spin = ExcitedState(lr_spin, i)
        den_spin = ex_spin.get_pseudo_density() * Bohr**3
        ddiff = gd.integrate(abs(den_vspin - den_spin))
        parprint('   density integral, difference=',
                 gd.integrate(den_vspin), gd.integrate(den_spin),
                 ddiff)
        parprint('   aed density integral',
                 finegd.integrate(ex_vspin.get_all_electron_density()
                                  * Bohr**3),
                 finegd.integrate(ex_spin.get_all_electron_density()
                                  * Bohr**3))
        assert ddiff < 3.e-3, ddiff


@pytest.mark.lrtddft
def test_io(in_tmp_dir, lr):
    fname = 'lr.dat.gz'
    lr.write(fname)
    world.barrier()

    lr2 = LrTDDFT.read(fname)
    lr2.diagonalize()

    equal(lr[0].get_energy(), lr2[0].get_energy(), 1e-6)




import numpy as np
from pathlib import Path
from gpaw.mpi import world, broadcast


def load_data(mml_name='mml.npz'):
    """
    Load the data and distribute among cores

    Input:
        mml_name        NLO data filename (default mml.npz)
    Output:

        p_kvnn          The mometum matrix elements, dimension (nk,3,nb,nb)
    """

    # Load the data to the master
    if world.rank == 0:
        nlo = np.load(mml_name)
        # print(nlo.files)
        # print(nlo['p_kvnn'])
    else:
        nlo = dict.fromkeys(['w_sk', 'f_skn', 'E_skn', 'p_skvnn'])

    # Distribute the data among cores
    k_info = distribute_data(
        [nlo['w_sk'], nlo['f_skn'], nlo['E_skn'], nlo['p_skvnn']])

    return k_info


def distribute_data(arr_list):
    """
    Distribute the data among the cores

    Input:
        arr_list        A list of numpy array (the first two should be s,k)
    Output:
        k_info          A  dictionary of data with key of s,k index
    """

    # Check the array shape
    size = world.size
    rank = world.rank
    if rank == 0:
        nk = 0
        arr_shape = []
        for ii, arr in enumerate(arr_list):
            ar_shape = arr.shape
            arr_shape.append(ar_shape)
            if nk == 0:
                ns = ar_shape[0]
                nk = ar_shape[1]
            else:
                assert ar_shape[1] == nk, 'Wrong shape for array.'
    else:
        arr_shape = None
        nk = None
        ns = None
    arr_shape = broadcast(arr_shape, root=0)
    nk = broadcast(nk, root=0)
    ns = broadcast(ns, root=0)

    # Distribute the data of k-points between cores
    k_info = {}

    # Loop over k points
    for s1 in range(ns):
        for kk in range(nk):
            if rank == 0:
                if kk % size == rank:
                    k_info[s1 * nk + kk] = [arr[s1, kk] for arr in arr_list]
                else:
                    for ii, arr in enumerate(arr_list):
                        data_k = np.array(arr[s1, kk], dtype=complex)
                        world.send(
                            data_k, kk % size, tag=ii * nk + kk)
            else:
                if kk % size == rank:
                    dataset = []
                    for ii, cshape in enumerate(arr_shape):
                        data_k = np.empty(cshape[2:], dtype=complex)
                        world.receive(data_k, 0, tag=ii * nk + kk)
                        dataset.append(data_k)
                    k_info[s1 * nk + kk] = dataset

    return k_info


def is_file(filename):
    """
    Check if the file exist

    Input:
        filename        Filename to check
    Output:
        file_exist      Flag for file
    """
    if world.rank == 0:
        if (not Path(filename).is_file()):
            file_exist = True
        else:
            file_exist = False
    else:
        file_exist = None

    file_exist = broadcast(file_exist, 0)

    return file_exist



from __future__ import annotations

import contextlib
import os
import sys
from pathlib import Path
from typing import IO

from gpaw.mpi import MPIComm, world
from gpaw.yml import obj2yaml as o2y


class Logger:
    def __init__(self,
                 filename: str | Path | IO[str] | None = '-',
                 comm: MPIComm | None = None):
        self.comm = comm or world

        self.fd: IO[str]

        if self.comm.rank > 0 or filename is None:
            self.fd = open(os.devnull, 'w', encoding='utf-8')
            self.close_fd = True
        elif filename == '-':
            self.fd = sys.stdout
            self.close_fd = False
        elif isinstance(filename, (str, Path)):
            self.fd = open(filename, 'w', encoding='utf-8')
            self.close_fd = True
        else:
            self.fd = filename
            self.close_fd = False

        self.indentation = ''

    def __del__(self) -> None:
        if self.close_fd:
            self.fd.close()

    @contextlib.contextmanager
    def indent(self, text):
        self(text)
        self.indentation += '  '
        yield
        self.indentation = self.indentation[2:]

    @contextlib.contextmanager
    def comment(self):
        self.indentation += '# '
        yield
        self.indentation = self.indentation[2:]

    def __call__(self, *args, **kwargs) -> None:
        if not self.fd.closed:
            i = self.indentation
            if kwargs:
                for kw, arg in kwargs.items():
                    assert kw not in ['end', 'sep', 'flush', 'file'], kw
                    print(f'{i}{kw}: {o2y(arg, i + "  ")}',
                          file=self.fd)
            else:
                text = ' '.join(str(arg) for arg in args)
                if i:
                    text = i + text.replace('\n', '\n' + i)
                print(text, file=self.fd)



# Copyright (C) 2008  CAMd
# Please see the accompanying LICENSE file for further information.

"""This module is used in delta self-consistent field (dSCF) calculations

dSCF is a simple 'ad hoc' method to estimating excitation energies within
DFT. The only difference to ordinary DFT is that one or more electrons(s)
are forced to occupy one or more predefined orbitals. The only restriction
on these orbitals is that they must be linear combinations of available
Kohn-Sham orbitals.

"""

import sys
import numpy as np

from ase.parallel import paropen
from gpaw.utilities import devnull
from ase.units import Ha

from gpaw.occupations import FermiDiracCalculator


def dscf_calculation(paw, orbitals, atoms):
    """Helper function to prepare a calculator for a dSCF calculation

    Parameters
    ==========
    orbitals: list of lists
        Orbitals which one wants to occupy. The format is
        orbitals = [[1.0,orb1,0],[1.0,orb2,1],...], where 1.0 is the no.
        of electrons, orb1 and orb2 are the orbitals (see MolecularOrbitals
        below for an example of an orbital class). 0 and 1 represents the
        spin (up and down). This number is ignored in a spin-paired
        calculation.

    Example
    =======

    ::

        atoms.calc = calc
        e_gs = atoms.get_potential_energy() #ground state energy
        sigma_star=MolecularOrbitals(calc, molecule=[0,1],
                                     w=[[1.,0.,0.,0.],[-1.,0.,0.,0.]])
        dscf_calculation(calc, [[1.0,sigma_star,1]], atoms)
        e_exc = atoms.get_potential_energy() #excitation energy

    """

    # If the calculator has not been initialized the occupation object
    # is None
    if paw.wfs is None:
        paw.initialize(atoms)
    assert paw.wfs.mode != 'pw'
    occ = paw.wfs.occupations
    if isinstance(occ, OccupationsDSCF):
        occ.orbitals = orbitals
    else:
        if occ.name == 'zero-width':
            occ = FermiDiracCalculator(1e-4, occ.parallel_layout)
        new_occ = OccupationsDSCF(occ, paw, orbitals)
        paw.wfs.occupations = new_occ
    # If the calculator has already converged (for the ground state),
    # reset self-consistency and let the density be updated right away
    if paw.scf.converged:
        paw.scf.niter_fixdensity = 0
        paw.scf.reset()


class OccupationsDSCF:
    """Occupation class.

    Corresponds to the ordinary FermiDirac class in occupation.py. Only
    difference is that it forces some electrons in the supplied orbitals
    in stead of placing all the electrons by a Fermi-Dirac distribution.
    """

    def __init__(self, occ, calc, orbitals):
        self.occ = occ
        self.extrapolate_factor = occ.extrapolate_factor
        self.wfs = calc.wfs
        self.orbitals = orbitals
        self.norbitals = len(self.orbitals)

        self.cnoe = 0.0
        for orb in self.orbitals:
            self.cnoe += orb[0]

    def calculate(self,
                  nelectrons,
                  eigenvalues,
                  weights,
                  fermi_levels_guess):
        f_qn, fermi_levels, e_entropy = self.occ.calculate(
            nelectrons - self.cnoe,
            eigenvalues,
            weights,
            fermi_levels_guess)

        # Get the expansion coefficients c_un for each dscf-orbital
        # and incorporate their respective occupations into kpt.ne_o
        c_oun = []
        for orb in self.orbitals:
            c_oun.append(orb[1].expand([e / Ha for e in fermi_levels],
                                       self.wfs,
                                       self.occ.todict().get('fixmagmom')))

        for u, kpt in enumerate(self.wfs.kpt_u):
            kpt.ne_o = np.zeros(self.norbitals, dtype=float)
            kpt.c_on = np.zeros((self.norbitals, len(kpt.eps_n)),
                                dtype=complex)

            for o, orb in enumerate(self.orbitals):
                # TODO XXX false if orb[0]<0 since abs(c_n)**2>0
                # kpt.c_on[o,:] = abs(orb[0])**0.5 * c_oun[o][u]

                kpt.ne_o[o] = orb[0]
                kpt.c_on[o, :] = c_oun[o][u]

                if self.wfs.nspins == 2:
                    assert orb[2] in range(2), 'Invalid spin index'

                    if orb[2] == kpt.s:
                        kpt.ne_o[o] *= kpt.weight
                    else:
                        kpt.ne_o[o] = 0.0
                else:
                    kpt.ne_o[o] *= 0.5 * kpt.weight

        return f_qn, fermi_levels, e_entropy

    def calculate_band_energy(self, wfs):
        de_band = 0.0
        for kpt in wfs.kpt_u:
            for ne, c_n in zip(kpt.ne_o, kpt.c_on):
                de_band += ne * np.dot(np.abs(c_n)**2, kpt.eps_n)
        return de_band


class MolecularOrbital:
    r"""Class defining orbitals that should be filled in a dSCF calculation.

    An orbital is defined through a linear combination of the atomic
    partial waves. In each self-consistent cycle the method expand
    is called. This method take the Kohn-Sham orbitals fulfilling the
    criteria given by Estart, Eend and nos and return the best
    possible expansion of the orbital in this basis. The integral
    of the Kohn-Sham all-electron wavefunction ``|u,n>`` (u being local spin
    and kpoint index) and the partial wave ``|\phi_i^a>`` is approximated
    by::

      wfs.kpt_u[u].P_ani = <\tilde p_i^a|\tilde\psi_{un}>.

    Parameters
    ----------
    paw: gpaw calculator instance
        The calculator used in the dSCF calculation.
    Estart: float
        Kohn-Sham orbitals with an energy above Efermi+Estart are used
        in the linear expansion.
    Eend: float
        Kohn-Sham orbitals with an energy below Efermi+Eend are used
        in the linear expansion.
    nos: int
        The maximum Number Of States used in the linear expansion.
    weights: dictionary
        The keys represent atoms and have values which are lists of weights
        of the contributing partial waves. The default value thuis corresponds
        to an antibonding 2\sigma orbital of atoms 0 and 1.
        Format::

          {1. atom: [weight of 1. projector function of the 1. atom,
                     weight of 2. projector function of the 1. atom, ...],
           2. atom: [weight of 1. projector function of the 2. atom,
                     weight of 2. projector function of the 2. atom, ...],
           ...}
    """

    def __init__(self, paw, Estart=0.0, Eend=1.e6,
                 nos=None, weights={0: [1], 1: [-1]}):

        self.w = weights
        self.Estart = Estart
        self.Eend = Eend
        self.nos = nos

    def expand(self, epsF, wfs, fixmom):
        if wfs.nspins == 2 and not fixmom:
            epsF = epsF * 2

        if self.nos is None:
            self.nos = wfs.bd.nbands

        c_un = []
        for u, kpt in enumerate(wfs.kpt_u):
            Porb_n = np.zeros(wfs.bd.nbands, dtype=complex)
            for a, P_ni in kpt.P_ani.items():
                if a in self.w.keys():
                    for i in range(len(self.w[a])):
                        Porb_n += self.w[a][i] * P_ni[:, i]
            wfs.gd.comm.sum(Porb_n)

            # Starting from KS orbitals with largest overlap,
            # fill in the expansion coeffients between Estart and Eend
            c_n = np.zeros(wfs.bd.nbands, dtype=complex)
            nos = 0
            bandpriority = np.argsort(abs(Porb_n)**2)[::-1]

            for n in bandpriority:
                if (kpt.eps_n[n] > epsF[kpt.s] + self.Estart and
                    kpt.eps_n[n] < epsF[kpt.s] + self.Eend):
                    c_n[n] = Porb_n[n].conj()
                    nos += 1
                if nos == self.nos:
                    break

            # Normalize expansion coefficients
            c_n /= np.sqrt(sum(abs(c_n)**2))
            c_un.append(c_n)

        return c_un


class AEOrbital:
    """Class defining the orbitals that should be filled in a dSCF calculation.

    An orbital is defined through a linear combination of KS orbitals
    which is determined by this class as follows: For each kpoint and spin
    we calculate the quantity ``c_n = <n|a>`` where ``|n>`` is the
    all-electron KS states in the calculation and ``|a>`` is the
    all-electron resonant state to be kept occupied. We can then
    write ``|a> = Sum(c_n|n>)`` and in each self-consistent cycle the
    method expand is called. This method take the Kohn-Sham
    orbitals fulfilling the criteria given by Estart, Eend and
    nos (Number Of States) and return the best possible expansion of
    the orbital in this basis.

    Parameters
    ----------
    paw: gpaw calculator instance
        The calculator used in the dSCF calculation.
    molecule: list of integers
        The atoms, which are a part of the molecule.
    Estart: float
        Kohn-Sham orbitals with an energy above Efermi+Estart are used
        in the linear expansion.
    Eend: float
        Kohn-Sham orbitals with an energy below Efermi+Eend are used
        in the linear expansion.
    nos: int
        The maximum Number Of States used in the linear expansion.
    wf_u: list of wavefunction arrays
        Wavefunction to be occupied on the kpts on this processor:

        wf_u = [kpt.psit_nG[n] for kpt in calc_mol.wfs.kpt_u]

    p_uai: list of dictionaries
        Projector overlaps with the wavefunction to be occupied for each
        kpoint. These are used when correcting to all-electron wavefunction
        overlaps:

        p_uai = [dict([(mol[a], P_ni[n]) for a, P_ni in kpt.P_ani.items()])
                 for kpt in paw.wfs.kpt_u]

        where mol is a list of atoms contributing to the state and n is the
        band. See examples in the dscf documentation on the gpaw webpage.
    """

    def __init__(self, paw, wf_u, p_uai, Estart=0.0, Eend=1.e6, nos=None,
                 txt='-'):

        self.wf_u = wf_u
        self.p_uai = p_uai
        self.Estart = Estart
        self.Eend = Eend
        self.nos = nos

        if txt is None:
            self.txt = devnull
        elif txt == '-':
            self.txt = sys.stdout
        elif isinstance(txt, str):
            self.txt = paropen(txt, 'w')
        else:
            self.txt = txt

    def expand(self, epsF, wfs, fixmom):

        if wfs.mode == 'lcao':
            wfs.initialize_wave_functions_from_lcao()
        if wfs.nspins == 2 and not fixmom:
            epsF = epsF * 2

        if self.nos is None:
            self.nos = wfs.bd.nbands

        # Check dimension of lists
        if len(self.wf_u) == len(wfs.kpt_u):
            wf_u = self.wf_u
            p_uai = self.p_uai
        else:
            raise RuntimeError('List of wavefunctions has wrong size')

        c_un = []
        p_u = []
        for u, kpt in enumerate(wfs.kpt_u):

            # Inner product of pseudowavefunctions
            wf = np.reshape(wf_u[u], -1)
            Wf_n = kpt.psit_nG
            Wf_n = np.reshape(Wf_n, (len(Wf_n), -1))
            Porb_n = np.dot(Wf_n.conj(), wf) * wfs.gd.dv

            # Correction to obtain inner product of AE wavefunctions
            for a, p_i in p_uai[u].items():
                for n in range(wfs.bd.nbands):
                    for i in range(len(p_i)):
                        for j in range(len(p_i)):
                            Porb_n[n] += (kpt.P_ani[a][n][i].conj() *
                                          wfs.setups[a].dO_ii[i][j] *
                                          p_i[j])
            wfs.gd.comm.sum(Porb_n)

            # print 'Kpt:', kpt.k, ' Spin:', kpt.s, \
            #      ' Sum_n|<orb|nks>|^2:', sum(abs(Porb_n)**2)
            p_u.append(np.array([sum(abs(Porb_n)**2)], dtype=float))

            # Starting from KS orbitals with largest overlap,
            # fill in the expansion coeffients
            c_n = np.zeros(wfs.bd.nbands, dtype=complex)
            nos = 0
            bandpriority = np.argsort(abs(Porb_n)**2)[::-1]

            for n in bandpriority:
                if (kpt.eps_n[n] > epsF[kpt.s] + self.Estart and
                    kpt.eps_n[n] < epsF[kpt.s] + self.Eend):
                    c_n[n] = Porb_n[n]
                    nos += 1
                if nos == self.nos:
                    break

            # Normalize expansion coefficients
            c_n /= np.sqrt(sum(abs(c_n)**2))
            c_un.append(c_n)

        for s in range(wfs.nspins):
            for k in range(wfs.kd.nibzkpts):
                p = wfs.collect_auxiliary(p_u, k, s)
                if wfs.world.rank == 0:
                    self.txt.write('Kpt: %d, Spin: %d, '
                                   'Sum_n|<orb|nks>|^2: %f\n' % (k, s, p))

        return c_un



from __future__ import annotations
from typing import Callable

import numpy as np

from gpaw.core.plane_waves import PWArray
from gpaw.core.uniform_grid import UGArray
from gpaw.core.arrays import DistributedArrays as XArray
from gpaw.gpu import cupy as cp
from gpaw.new import zips
from gpaw.new.hamiltonian import Hamiltonian
from gpaw.new.c import pw_precond


class PWHamiltonian(Hamiltonian):
    def __init__(self, grid, pw, xp):
        self.plan = grid.new(dtype=pw.dtype).fft_plans(xp=xp)
        self.pw_cache = {}

    def apply_local_potential(self,
                              vt_R: UGArray,
                              psit_nG: XArray,
                              out: XArray) -> None:
        assert isinstance(psit_nG, PWArray)
        assert isinstance(out, PWArray)
        out_nG = out
        xp = psit_nG.xp
        pw = psit_nG.desc
        if xp is not np and pw.comm.size == 1 and pw.dtype == complex:
            return apply_local_potential_gpu(vt_R, psit_nG, out_nG)
        vt_R = vt_R.gather(broadcast=True)
        grid = vt_R.desc.new(comm=None, dtype=psit_nG.desc.dtype)
        tmp_R = grid.empty(xp=xp)
        if pw.comm.size == 1:
            pw_local = pw
        else:
            key = tuple(pw.kpt_c)
            pw_local = self.pw_cache.get(key)
            if pw_local is None:
                pw_local = pw.new(comm=None)
                self.pw_cache[key] = pw_local
        psit_G = pw_local.empty(xp=xp)
        e_kin_G = xp.asarray(psit_G.desc.ekin_G)
        domain_comm = psit_nG.desc.comm
        mynbands = psit_nG.mydims[0]
        vtpsit_G = pw_local.empty(xp=xp)
        for n1 in range(0, mynbands, domain_comm.size):
            n2 = min(n1 + domain_comm.size, mynbands)
            psit_nG[n1:n2].gather_all(psit_G)
            if domain_comm.rank < n2 - n1:
                psit_G.ifft(out=tmp_R)
                tmp_R.data *= vt_R.data
                tmp_R.fft(out=vtpsit_G)
                psit_G.data *= e_kin_G
                vtpsit_G.data += psit_G.data
            out_nG[n1:n2].scatter_from_all(vtpsit_G)

    def apply_mgga(self,
                   dedtaut_R: UGArray,
                   psit_nG: XArray,
                   vt_nG: XArray) -> None:
        pw = psit_nG.desc
        dpsit_R = dedtaut_R.desc.new(dtype=pw.dtype).empty()
        Gplusk1_Gv = pw.reciprocal_vectors()
        tmp_G = pw.empty()

        for psit_G, vt_G in zips(psit_nG, vt_nG):
            for v in range(3):
                tmp_G.data[:] = psit_G.data
                tmp_G.data *= 1j * Gplusk1_Gv[:, v]
                tmp_G.ifft(out=dpsit_R)
                dpsit_R.data *= dedtaut_R.data
                dpsit_R.fft(out=tmp_G)
                vt_G.data -= 0.5j * Gplusk1_Gv[:, v] * tmp_G.data

    def create_preconditioner(self,
                              blocksize: int
                              ) -> Callable[[PWArray,
                                             PWArray,
                                             PWArray], None]:
        return precondition


def precondition(psit_nG: PWArray,
                 residual_nG: PWArray,
                 out: PWArray) -> None:
    """Preconditioner for KS equation.

    From:

      Teter, Payne and Allen, Phys. Rev. B 40, 12255 (1989)

    as modified by:

      Kresse and Furthmüller, Phys. Rev. B 54, 11169 (1996)
    """

    xp = psit_nG.xp
    G2_G = xp.asarray(psit_nG.desc.ekin_G * 2)
    ekin_n = psit_nG.norm2('kinetic')

    if xp is np:
        for r_G, o_G, ekin in zips(residual_nG.data,
                                   out.data,
                                   ekin_n):
            pw_precond(G2_G, r_G, ekin, o_G)
        return

    out.data[:] = gpu_prec(ekin_n[:, np.newaxis],
                           G2_G[np.newaxis],
                           residual_nG.data)


@cp.fuse()
def gpu_prec(ekin, G2, residual):
    x = 1 / ekin / 3 * G2
    a = 27.0 + x * (18.0 + x * (12.0 + x * 8.0))
    xx = x * x
    return -4.0 / 3 / ekin * a / (a + 16.0 * xx * xx) * residual


def spinor_precondition(psit_nsG, residual_nsG, out):
    G2_G = psit_nsG.desc.ekin_G * 2
    for r_sG, o_sG, ekin in zips(residual_nsG.data,
                                 out.data,
                                 psit_nsG.norm2('kinetic').sum(1)):
        for r_G, o_G in zips(r_sG, o_sG):
            pw_precond(G2_G, r_G, ekin, o_G)


class SpinorPWHamiltonian(Hamiltonian):
    def apply(self,
              vt_xR: UGArray,
              dedtaut_xR: UGArray | None,
              psit_nsG: XArray,
              out: XArray,
              spin: int) -> XArray:
        assert dedtaut_xR is None
        out_nsG = out
        pw = psit_nsG.desc

        if pw.qspiral_v is None:
            np.multiply(pw.ekin_G, psit_nsG.data, out_nsG.data)
        else:
            for s, sign in enumerate([1, -1]):
                ekin_G = 0.5 * ((pw.G_plus_k_Gv +
                                 0.5 * sign * pw.qspiral_v)**2).sum(1)
                np.multiply(ekin_G, psit_nsG.data[:, s], out_nsG.data[:, s])

        grid = vt_xR.desc.new(dtype=complex)

        v, x, y, z = vt_xR.data
        iy = y * 1j

        f_sR = grid.empty(2)
        g_R = grid.empty()

        for p_sG, o_sG in zips(psit_nsG, out_nsG):
            p_sG.ifft(out=f_sR)
            a, b = f_sR.data
            g_R.data = a * (v + z) + b * (x - iy)
            o_sG.data[0] += g_R.fft(pw=pw).data
            g_R.data = a * (x + iy) + b * (v - z)
            o_sG.data[1] += g_R.fft(pw=pw).data

        return out_nsG

    def create_preconditioner(self, blocksize):
        return spinor_precondition


def apply_local_potential_gpu(vt_R, psit_nG, out_nG):
    from gpaw.gpu import cupyx
    pw = psit_nG.desc
    e_kin_G = cp.asarray(pw.ekin_G)
    mynbands = psit_nG.mydims[0]
    plan = vt_R.desc.fft_plans(xp=cp, dtype=complex)
    Q_G = plan.indices(pw)
    shape = tuple(vt_R.desc.size_c)
    blocksize = 10
    psit_bR = None
    for b1 in range(0, mynbands, blocksize):
        b2 = min(b1 + blocksize, mynbands)
        nb = b2 - b1
        if psit_bR is None:
            psit_bR = cp.empty((nb,) + shape, complex)
        elif nb < blocksize:
            psit_bR = psit_bR[:nb]
        psit_bR[:] = 0.0
        psit_bR.reshape((nb, -1))[:, Q_G] = psit_nG.data[b1:b2]
        psit_bR[:] = cupyx.scipy.fft.ifftn(
            psit_bR,
            shape,
            norm='forward',
            overwrite_x=True)
        psit_bR *= vt_R.data
        psit_bR[:] = cupyx.scipy.fft.fftn(
            psit_bR,
            shape,
            norm='forward',
            overwrite_x=True)
        out_nG.data[b1:b2] = psit_nG.data[b1:b2]
        out_nG.data[b1:b2] *= e_kin_G
        out_nG.data[b1:b2] += psit_bR.reshape((nb, -1))[:, Q_G]



"""Test EXX/HFT implementation."""
import pytest
from ase import Atoms
from gpaw import GPAW
from gpaw.test import equal
from gpaw.xc import XC
from gpaw.xc.hybrid import HybridXC


def xc1(name):
    return dict(name=name, stencil=1)


@pytest.mark.libxc
@pytest.mark.hybrids
def test_exx_exx(in_tmp_dir):
    be2 = Atoms('Be2', [(0, 0, 0), (2.45, 0, 0)])
    be2.center(vacuum=2.0)

    ref_1871 = {  # Values from revision 1871. Not true reference values
        # xc         Energy          eigenvalue 0    eigenvalue 1
        'PBE': (5.424066548470926, -3.84092, -0.96192),
        'PBE0': (-790.919942, -4.92321, -1.62948),
        'EXX': (-785.5837828306236, -7.16802337336, -2.72602997017)}

    current = {}  # Current revision
    for xc in [XC(xc1('PBE')),
               HybridXC('PBE0', stencil=1, finegrid=True),
               HybridXC('EXX', stencil=1, finegrid=True),
               XC(xc1('PBE'))]:  # , 'oldPBE', 'LDA']:
        # Generate setup
        # g = Generator('Be', setup, scalarrel=True, nofiles=True, txt=None)
        # g.run(exx=True, **parameters['Be'])

        # switch to new xc functional
        calc = GPAW(mode='fd',
                    xc=xc,
                    h=0.21,
                    eigensolver='rmm-diis',
                    nbands=3,
                    convergence={'eigenstates': 1e-6},
                    txt='exx.txt')
        be2.calc = calc
        E = be2.get_potential_energy()
        if xc.name != 'PBE':
            E += calc.get_reference_energy()
        bands = calc.get_eigenvalues()[:2]  # not 3 as unocc. eig are random!?
        res = (E,) + tuple(bands)
        print(xc.name, res)

        if xc.name in current:
            for first, second in zip(current[xc.name], res):
                equal(first, second, 2.5e-3)
        else:
            current[xc.name] = res

    for name in current:
        for ref, cur in zip(ref_1871[name], current[name]):
            print(ref, cur, ref - cur)
            equal(ref, cur, 2.9e-3)



"""Test poisson solver for asymmetric charges."""
from gpaw.utilities.gauss import Gaussian
from gpaw.grid_descriptor import GridDescriptor
from gpaw.poisson import PoissonSolver


# Initialize classes

def test_poisson_poisson_asym():
    a = 20.0  # Size of cell
    inv_width = 19  # inverse width of the gaussian
    N = 48  # Number of grid points
    center_of_charge = (a / 2, a / 2, 3 * a / 4)  # off center charge
    Nc = (N, N, N)                # Number of grid points along each axis
    gd = GridDescriptor(Nc, (a, a, a), 0)    # Grid-descriptor object
    solver = PoissonSolver(nn=3, use_charge_center=True)
    solver.set_grid_descriptor(gd)
    gauss = Gaussian(gd, a=inv_width, center=center_of_charge)
    test_poisson = Gaussian(gd, a=inv_width, center=center_of_charge)

    # Check if Gaussian potentials are made correctly

    # Array for storing the potential
    pot = gd.zeros(dtype=float, global_array=False)
    solver.load_gauss()
    vg = test_poisson.get_gauss_pot(0)
    # Get analytic functions
    ng = gauss.get_gauss(0)
    #    vg = solver.phi_gauss
    # Solve potential numerically
    solver.solve(pot, ng, charge=1.0, zero_initial_phi=False)
    # Determine residual
    # residual = norm(pot - vg)
    residual = gd.integrate((pot - vg)**2)**0.5

    print('residual %s' % (
        residual))
    assert residual < 1e-5  # Better than 5.x

    # mpirun -np 2 python gauss_func.py --gpaw-parallel --gpaw-debug



from typing import Any

import numpy as np
from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator


def obj2yaml(obj: Any, indentation: str = '') -> str:
    """Convert Python object to YAML string.

    >>> print(obj2yaml({'a': {'b': 42}}))
    a:
      b: 42
    """
    if isinstance(obj, dict):
        i = indentation
        txt = f'\n{i}'.join(f'{k}: {obj2yaml(v, i + "  ")}'
                            for k, v in obj.items())
        if i:
            return '\n' + i + txt
        return txt.replace(': \n', ':\n')
    return repr(obj)


def indent(text: Any, indentation='  ') -> str:
    r"""Indent text blob.

    >>> indent('line 1\nline 2', '..')
    '..line 1\n..line 2'
    """
    if not isinstance(text, str):
        text = str(text)
    return indentation + text.replace('\n', '\n' + indentation)


def comment(text: Any) -> str:
    """Comment text blob.

    >>> comment('hmm')
    '# hmm'
    """
    return indent(text, '# ')


def read_gpaw_yaml(fd, index):
    r"""Convert YAML text output from GPAW calculation to Atoms object(s).

    >>> import io
    >>> y = ('atoms: [[H, [0, 0, 0], [0, 0, 1]]]\n'
    ...      'cell: [[0, 0, 0], [0, 0, 0], [0, 0, 0]]\n'
    ...      'periodic: [false, false, false]\n')
    >>> read_gpaw_yaml(io.StringIO(y), 0)
    Atoms(symbols='H', pbc=False)
    """
    import yaml
    configs = []
    for dct in yaml.safe_load_all(fd):
        if 'atoms' in dct:
            atoms = dict2atoms(dct)
            configs.append(atoms)
    return configs[index]


def dict2atoms(dct) -> Atoms:
    symbols = []
    positions = []
    magmoms = []
    for symbol, position, (_, _, magmom) in dct['atoms']:
        symbols.append(symbol)
        positions.append(position)
        magmoms.append(magmom)

    cell = dct['cell']
    pbc = dct['periodic']

    atoms = Atoms(symbols,
                  positions,
                  cell=cell,
                  pbc=pbc)

    if 'energies' in dct:
        energy = dct['energies']['extrapolated']
        if 'forces' in dct:
            forces = dct['forces']
        else:
            forces = None
        if 'stress tensor' in dct:
            stress = np.array(dct['stress tensor']).ravel()[0, 4, 8, 5, 2, 1]
        else:
            stress = None
        atoms.calc = SinglePointCalculator(energy=energy,
                                           forces=forces,
                                           stress=stress,
                                           atoms=atoms)
    return atoms


if __name__ == '__main__':
    import sys
    import pprint
    import yaml
    for dct in yaml.safe_load_all(open(sys.argv[1])):
        print('---')
        pprint.pp(dct)



import numpy as np
from scipy.special import erf
from ase.units import Bohr


class DipoleCorrection:
    """Dipole-correcting wrapper around another PoissonSolver."""
    def __init__(self, poissonsolver, direction, width=1.0):
        """Construct dipole correction object.

        poissonsolver:
            Poisson solver.
        direction: int or str
            Specification of layer: 0, 1, 2, 'xy', 'xz' or 'yz'.
        width: float
            Width in Angstrom of dipole layer used for the plane-wave
            implementation.
        """
        self.c = direction
        self.poissonsolver = poissonsolver
        self.width = width / Bohr

        self.correction = None  # shift in potential
        self.sawtooth_q = None  # Fourier transformed sawtooth

    def todict(self):
        dct = self.poissonsolver.todict()
        dct['dipolelayer'] = self.c
        if self.width != 1.0 / Bohr:
            dct['width'] = self.width * Bohr
        return dct

    def get_stencil(self):
        return self.poissonsolver.get_stencil()

    def set_grid_descriptor(self, gd):
        self.poissonsolver.set_grid_descriptor(gd)
        self.check_direction(gd, gd.pbc_c)

    def check_direction(self, gd, pbc_c):
        if isinstance(self.c, str):
            axes = ['xyz'.index(d) for d in self.c]
            for c in range(3):
                if abs(gd.cell_cv[c, axes]).max() < 1e-12:
                    break
            else:
                raise ValueError('No axis perpendicular to {0}-plane!'
                                 .format(self.c))
            self.c = c

        if pbc_c[self.c]:
            raise ValueError('System must be non-periodic perpendicular '
                             'to dipole-layer.')

        # Right now the dipole correction must be along one coordinate
        # axis and orthogonal to the two others.  The two others need not
        # be orthogonal to each other.
        for c1 in range(3):
            if c1 != self.c:
                if abs(np.dot(gd.cell_cv[self.c], gd.cell_cv[c1])) > 1e-12:
                    raise ValueError('Dipole correction axis must be '
                                     'orthogonal to the two other axes.')

    def get_description(self):
        poissondesc = self.poissonsolver.get_description()
        desc = 'Dipole correction along %s-axis' % 'xyz'[self.c]
        return '\n'.join([poissondesc, desc])

    def initialize(self):
        self.poissonsolver.initialize()

    def solve(self, pot, dens, **kwargs):
        # Note that fdsolve() returns number of iterations and pwsolve()
        # returns the energy!!  This is because the
        # ChargedReciprocalSpacePoissonSolver has corrections to
        # the energy ...
        if isinstance(dens, np.ndarray):
            # finite-diference Poisson solver:
            return self.fdsolve(pot, dens, **kwargs)
        # Plane-wave solver:
        return self.pwsolve(pot, dens)

    def fdsolve(self, vHt_g, rhot_g, **kwargs):
        gd = self.poissonsolver.gd
        drhot_g, dvHt_g, self.correction = dipole_correction(
            self.c, gd, rhot_g)
        vHt_g -= dvHt_g
        iters = self.poissonsolver.solve(vHt_g, rhot_g + drhot_g, **kwargs)
        vHt_g += dvHt_g
        return iters

    def pwsolve(self, vHt_q, dens):
        gd = self.poissonsolver.pd.gd

        if self.sawtooth_q is None:
            self.initialize_sawtooth()

        epot = self.poissonsolver.solve(vHt_q, dens)

        dip_v = dens.calculate_dipole_moment()
        c = self.c
        L = gd.cell_cv[c, c]
        self.correction = 2 * np.pi * dip_v[c] * L / gd.volume
        vHt_q -= 2 * self.correction * self.sawtooth_q

        return epot + 2 * np.pi * dip_v[c]**2 / gd.volume

    def initialize_sawtooth(self):
        gd = self.poissonsolver.pd.gd
        if gd.comm.rank == 0:
            c = self.c
            L = gd.cell_cv[c, c]
            w = self.width / 2
            assert w < L / 2
            gc = int(w / gd.h_cv[c, c])
            x = gd.coords(c)
            sawtooth = x / L - 0.5
            a = 1 / L - 0.75 / w
            b = 0.25 / w**3
            sawtooth[:gc] = x[:gc] * (a + b * x[:gc]**2)
            sawtooth[-gc:] = -sawtooth[gc:0:-1]
            sawtooth_g = gd.empty(global_array=True)
            shape = [1, 1, 1]
            shape[c] = -1
            sawtooth_g[:] = sawtooth.reshape(shape)
            sawtooth_q = self.poissonsolver.pd.fft(sawtooth_g, local=True)
        else:
            sawtooth_q = None
        self.sawtooth_q = self.poissonsolver.pd.scatter(sawtooth_q)

    def estimate_memory(self, mem):
        self.poissonsolver.estimate_memory(mem)


def dipole_correction(c, gd, rhot_g, center=False, origin_c=None):
    """Get dipole corrections to charge and potential.

    Returns arrays drhot_g and dphit_g such that if rhot_g has the
    potential phit_g, then rhot_g + drhot_g has the potential
    phit_g + dphit_g, where dphit_g is an error function.

    The error function is chosen so as to be largely constant at the
    cell boundaries and beyond.
    """
    # This implementation is not particularly economical memory-wise

    moment = gd.calculate_dipole_moment(rhot_g, center=center,
                                        origin_c=origin_c)[c]
    if abs(moment) < 1e-12:
        return gd.zeros(), gd.zeros(), 0.0

    r_g = gd.get_grid_point_coordinates()[c]
    cellsize = abs(gd.cell_cv[c, c])
    sr_g = 2.0 / cellsize * r_g - 1.0  # sr ~ 'scaled r'
    alpha = 12.0  # should perhaps be variable
    drho_g = sr_g * np.exp(-alpha * sr_g**2)
    moment2 = gd.calculate_dipole_moment(drho_g, center=center,
                                         origin_c=origin_c)[c]
    factor = -moment / moment2
    drho_g *= factor
    phifactor = factor * (np.pi / alpha)**1.5 * cellsize**2 / 4.0
    dphi_g = -phifactor * erf(sr_g * np.sqrt(alpha))
    return drho_g, dphi_g, phifactor



from gpaw.symmetry import map_k_points_fast, map_k_points
from ase.dft.kpoints import monkhorst_pack
from gpaw import GPAW
from ase.build import bulk
import numpy as np
from itertools import product


def test_symmetry_kpoint_mapping():
    def test_mapping(bz2bz_ks, U_scc, bzk_kc, time_reversal=False):
        eps = 1e-7

        if time_reversal:
            U_scc = np.concatenate([U_scc, -U_scc])

        for k_c, bz2bz_s in zip(bzk_kc, bz2bz_ks):
            delta_sc = (np.dot(U_scc, k_c) -
                        bzk_kc[bz2bz_s, :])[bz2bz_s >= 0]
            delta_sc = np.abs(delta_sc - delta_sc.round())
            assert delta_sc.max() < eps

    # Test kpoint mapping functionality of the gpaw.symmetry module
    atoms = bulk('C')
    calc = GPAW(mode='pw',
                kpts={'size': (5, 5, 5)},
                txt=None)
    atoms.calc = calc
    atoms.get_potential_energy()

    U_scc = calc.wfs.kd.symmetry.op_scc
    time_reversal = False

    for gamma in [True, False]:
        for time_reversal in [True, False]:
            for i, j, k in product(*([range(1, 7)] * 3)):

                bzk_kc = monkhorst_pack((i, j, k))

                if gamma:
                    offset = (((i + 1) % 2) / (2 * i),
                              ((j + 1) % 2) / (2 * j),
                              ((k + 1) % 2) / (2 * k))
                    bzk_kc += offset

                bz2bz_ks = map_k_points(bzk_kc, U_scc,
                                        time_reversal,
                                        None, calc.wfs.kd.symmetry.tol)
                bz2bzfast_ks = map_k_points_fast(bzk_kc, U_scc,
                                                 time_reversal,
                                                 None,
                                                 calc.wfs.kd.symmetry.tol)

                assert ((bz2bz_ks - bz2bzfast_ks)**2 < 1e-9).all()

                test_mapping(bz2bz_ks, U_scc, bzk_kc, time_reversal)
                test_mapping(bz2bzfast_ks, U_scc, bzk_kc, time_reversal)



import pytest
from ase import Atoms
from ase.optimize import BFGS

from gpaw import PW, SCIPY_VERSION
from gpaw.new.ase_interface import GPAW as NewGPAW
from gpaw.calculator import GPAW as OldGPAW
from gpaw.mpi import world


@pytest.mark.parametrize(
    'gpu, GPAW',
    [(False, OldGPAW),
     (False, NewGPAW),
     pytest.param(True, NewGPAW, marks=pytest.mark.gpu)])
def test_pw_slab(gpu, GPAW):
    if GPAW is NewGPAW and world.size > 1:
        pytest.skip()
    if gpu and SCIPY_VERSION < [1, 6]:
        pytest.skip()

    a = 2.65
    slab = Atoms('Li2',
                 [(0, 0, 0), (0, 0, a)],
                 cell=(a, a, 3 * a),
                 pbc=True)
    k = 4
    parallel = {'band': min(world.size, 4)}
    if gpu:
        parallel['gpu'] = True
    calc = GPAW(mode=PW(200),
                eigensolver='rmm-diis',
                parallel=parallel,
                kpts=(k, k, 1))
    slab.calc = calc
    BFGS(slab).run(fmax=0.01)
    assert abs(slab.get_distance(0, 1) - 2.46) < 0.01



"""Test of HGH pseudopotential implementation.
This is the canonical makes-sure-nothing-breaks test, which checks that the
numbers do not change from whatever they were before.

The test runs an HGH calculation on a misconfigured H2O molecule, such that
the forces are nonzero.

Energy is compared to a previous calculation; if it differs significantly,
that is considered an error.

Forces are compared to a previous finite-difference result.
"""


import numpy as np
from ase.build import molecule

from gpaw import GPAW, PoissonSolver
from gpaw.test import equal


def test_pseudopotential_hgh_h2o():
    mol = molecule('H2O')
    mol.rattle(0.2)
    mol.center(vacuum=2.0)
    calc = GPAW(mode='fd',
                nbands=6,
                poissonsolver=PoissonSolver('fd'),
                gpts=(32, 40, 40),
                setups='hgh',
                convergence=dict(eigenstates=1e-9,
                                 density=1e-5,
                                 energy=0.3e-5),
                txt='-')
    mol.calc = calc
    e = mol.get_potential_energy()
    F_ac = mol.get_forces()

    F_ac_ref = np.array([[7.33077718, 3.81069249, -6.07405156],
                         [-0.9079617, -1.18203514, 3.43777589],
                         [-0.61642527, -0.41889306, 2.332415]])

    eref = -468.527121304

    eerr = abs(e - eref)

    print('energy', e)
    print('ref energy', eref)
    print('energy error', eerr)

    print('forces')
    print(F_ac)

    print('ref forces')
    print(F_ac_ref)

    ferr = np.abs(F_ac - F_ac_ref).max()
    print('max force error', ferr)

    fdcheck = False

    if fdcheck:
        from ase.calculators.test import numeric_forces
        F_ac_fd = numeric_forces(mol)

        print('Self-consistent forces')
        print(F_ac)
        print('FD forces')
        print(F_ac_fd)
        print()
        print(repr(F_ac_fd))
        print()
        err = np.abs(F_ac - F_ac_fd).max()
        print('max err', err)

    wfs = calc.wfs
    gd = wfs.gd
    psit_nG = wfs.kpt_u[0].psit_nG
    try:
        dH_asp = calc.hamiltonian.dH_asp
    except AttributeError:
        from gpaw.utilities import pack2
        dH_asii = calc.calculation.state.potential.dH_asii
        dH_asp = {a: pack2(dH_sii) for a, dH_sii in dH_asii.items()}

    assert eerr < 1e-3, 'energy changed from reference'
    assert ferr < 0.015, 'forces do not match FD check'

    # Sanity check.  In HGH, the atomic Hamiltonian is constant.
    # Also the projectors should be normalized
    for a, dH_sp in dH_asp.items():
        dH_p = dH_sp[0]
        K_p = wfs.setups[a].K_p
        # B_ii = wfs.setups[a].B_ii
        # assert np.abs(B_ii.diagonal() - 1).max() < 1e-3
        # print 'B_ii'
        # print wfs.setups[a].B_ii
        # Actually, H2O might not be such a good test, since there'll only
        # be one element in the atomic Hamiltonian for O and zero for H.
        # print 'dH_p', dH_p
        # print 'K_p', K_p

        assert np.abs(dH_p - K_p).max() < 1e-10, 'atomic Hamiltonian changed'

        # h_ii = setup.data.h_ii
        # print 'h_ii', h_ii
        # print 'dH_ii', dH_ii

    # Sanity check: HGH is normconserving
    for psit_G in psit_nG:
        norm = gd.integrate(psit_G**2)  # around 1e-15 !  Surprisingly good.
        assert abs(1 - norm) < 1e-10, 'Not normconserving'

    energy_tolerance = 0.0003
    equal(e, eref, energy_tolerance)



"""This module defines a linear response TDDFT-class."""
import numbers
import sys
from math import sqrt
from typing import Dict, Any
import numpy as np

from ase.units import Hartree
from ase.utils.timing import Timer

import gpaw.mpi as mpi
from gpaw.xc import XC
from gpaw.lrtddft.excitation import Excitation, ExcitationList, get_filehandle
from gpaw.lrtddft.kssingle import KSSingles
from gpaw.lrtddft.omega_matrix import OmegaMatrix
from gpaw.lrtddft.apmb import ApmB
from gpaw.lrtddft.spectrum import spectrum

__all__ = ['LrTDDFT', 'photoabsorption_spectrum', 'spectrum']


class LrTDDFT(ExcitationList):
    """Linear Response TDDFT excitation list class

    Input parameters:

    calculator:
    the calculator object after a ground state calculation

    nspins:
    number of spins considered in the calculation
    Note: Valid only for unpolarised ground state calculation

    eps:
    Minimal occupation difference for a transition (default 0.001)

    istart:
    First occupied state to consider
    jend:
    Last unoccupied state to consider

    xc:
    Exchange-Correlation approximation in the Kernel
    derivative_level:
    0: use Exc, 1: use vxc, 2: use fxc  if available

    filename:
    read from a file
    """

    default_parameters: Dict[str, Any] = {
        'nspins': None,
        'restrict': {},
        'xc': None,
        'derivative_level': 1,
        'numscale': 0.00001,
        'filename': None,
        'finegrid': 2,
        'force_ApmB': False,  # for tests
        'eh_comm': None,  # parallelization over eh-pairs
        'poisson': None}  # use calculator's Poisson

    def __init__(self, calculator=None, log=None, txt='-', **kwargs):

        self.energy_to_eV_scale = Hartree
        self.timer = Timer()
        self.diagonalized = False

        self.set(**kwargs)
        self.calculator = calculator

        ExcitationList.__init__(self, log=log, txt=txt)

        if self.eh_comm is None:
            self.eh_comm = mpi.serial_comm
        elif isinstance(self.eh_comm, (mpi.world.__class__,
                                       mpi.serial_comm.__class__)):
            # Correct type already.
            pass
        else:
            # world should be a list of ranks:
            self.eh_comm = mpi.world.new_communicator(
                np.asarray(self.eh_comm))

        if calculator is not None and calculator.initialized:
            # XXXX not ready for k-points
            assert len(calculator.wfs.kd.bzk_kc) == 1
            if calculator.wfs.mode not in ['fd', 'lcao']:
                raise RuntimeError('LrTDDFT supports only fd and lcao modes')
            if calculator.wfs.mode != 'lcao':
                calculator.converge_wave_functions()
            if calculator.density.nct_G is None:
                spos_ac = calculator.initialize_positions()
                calculator.wfs.initialize(calculator.density,
                                          calculator.hamiltonian, spos_ac)

            self.forced_update()

    @property
    def calculator(self):
        return self._calc

    @calculator.setter
    def calculator(self, calc):
        self._calc = calc

        if self.xc is None and calc is not None:
            if calc.initialized:
                self.xc = calc.hamiltonian.xc
            else:
                self.xc = calc.parameters['xc']

    def set(self, **kwargs):
        """Change parameters."""
        changed = []
        for key, value in LrTDDFT.default_parameters.items():
            if hasattr(self, key):
                value = getattr(self, key)  # do not overwrite
            setattr(self, key, kwargs.pop(key, value))
            if value != getattr(self, key):
                changed.append(key)

        for key in kwargs:
            raise KeyError('Unknown key ' + key)

        return changed

    def analyse(self, what=None, out=None, min=0.1):
        """Print info about the transitions.

        Parameters:
          1. what: I list of excitation indicees, None means all
          2. out : I where to send the output, None means sys.stdout
          3. min : I minimal contribution to list (0<min<1)
        """
        if what is None:
            what = range(len(self))
        elif isinstance(what, numbers.Integral):
            what = [what]

        if out is None:
            out = sys.stdout

        for i in what:
            print(str(i) + ':', self[i].analyse(min=min), file=out)

    def calculate(self, atoms):
        self.calculator = atoms.calc
        if not hasattr(self, 'Om') or self.calculator.check_state(atoms):
            self.calculator.get_potential_energy(atoms)
            self.forced_update()
        return self

    def forced_update(self):
        """Recalc yourself."""
        if not self.force_ApmB:
            Om = OmegaMatrix
            name = 'LrTDDFT'
            if self.xc:
                if isinstance(self.xc, str):
                    xc = XC(self.xc)
                else:
                    xc = self.xc
                if hasattr(xc, 'hybrid') and xc.hybrid > 0.0:
                    Om = ApmB
                    name = 'LrTDDFThyb'
        else:
            Om = ApmB
            name = 'LrTDDFThyb'

        kss = KSSingles(restrict=self.restrict,
                        log=self.log)
        atoms = self.calculator.get_atoms()
        kss.calculate(atoms, self.nspins)

        self.Om = Om(self.calculator, kss,
                     self.xc, self.derivative_level, self.numscale,
                     finegrid=self.finegrid, eh_comm=self.eh_comm,
                     poisson=self.poisson, log=self.log)
        self.name = name

    def diagonalize(self, **kwargs):
        """Diagonalize and save new Eigenvalues and Eigenvectors"""
        self.set(**kwargs)
        self.timer.start('diagonalize')
        self.timer.start('omega')
        self.Om.diagonalize(kwargs.pop('restrict', {}))
        self.timer.stop('omega')
        self.diagonalized = True

        # remove old stuff
        self.timer.start('clean')
        while len(self):
            self.pop()
        self.timer.stop('clean')

        self.log('LrTDDFT digonalized:')
        self.timer.start('build')
        for j in range(len(self.Om.kss)):
            self.append(LrTDDFTExcitation(self.Om, j))
            self.log(' ', str(self[-1]))
        self.timer.stop('build')
        self.timer.stop('diagonalize')

    @classmethod
    def read(cls, filename=None, fh=None, restrict={}, log=None, txt=None):
        """Read myself from a file"""
        lr = cls(log=log, txt=txt)
        timer = lr.timer
        timer.start('name')
        if fh is None:
            f = get_filehandle(lr, filename)
        else:
            f = fh
        timer.stop('name')

        timer.start('header')
        # get my name
        s = f.readline().strip()
        lr.name = s.split()[1]

        lr.xc = XC(f.readline().strip().split()[0])
        values = f.readline().split()
        eps = float(values[0])
        if len(values) > 1:
            lr.derivative_level = int(values[1])
            lr.numscale = float(values[2])
            lr.finegrid = int(values[3])
        else:
            # old writing style, use old defaults
            lr.numscale = 0.001
        timer.stop('header')

        timer.start('init_kss')
        kss = KSSingles.read(fh=f, log=log)
        assert eps == kss.restrict['eps']
        lr.restrict = kss.restrict.values
        timer.stop('init_kss')
        timer.start('init_obj')
        if lr.name == 'LrTDDFT':
            lr.Om = OmegaMatrix(kss=kss, filehandle=f, log=lr.log)
        else:
            lr.Om = ApmB(kss=kss, filehandle=f, log=lr.log)
        timer.stop('init_obj')

        if not len(restrict):
            timer.start('read diagonalized')
            # check if already diagonalized
            p = f.tell()
            s = f.readline()
            if s != '# Eigenvalues\n' or len(restrict):
                # no further info or selection of
                # Kohn-Sham states changed
                # go back to previous position
                f.seek(p)
            else:
                lr.diagonalized = True
                # load the eigenvalues
                n = int(f.readline().split()[0])
                for i in range(n):
                    lr.append(LrTDDFTExcitation(string=f.readline()))
                # load the eigenvectors
                timer.start('read eigenvectors')
                f.readline()
                for i in range(n):
                    lr[i].f = np.array([float(x) for x in
                                        f.readline().split()])
                    lr[i].kss = lr.kss
                timer.stop('read eigenvectors')
            timer.stop('read diagonalized')
        else:
            timer.start('diagonalize')
            lr.diagonalize(restrict=restrict)
            timer.stop('diagonalize')

        if fh is None:
            f.close()

        return lr

    @property
    def kss(self):
        return self.Om.kss

    def singlets_triplets(self):
        """Split yourself into a singlet and triplet object"""

        slr = LrTDDFT(nspins=self.nspins, xc=self.xc,
                      restrict=self.kss.restrict.values,
                      derivative_level=self.derivative_level,
                      numscale=self.numscale)
        tlr = LrTDDFT(nspins=self.nspins, xc=self.xc,
                      restrict=self.kss.restrict.values,
                      derivative_level=self.derivative_level,
                      numscale=self.numscale)
        slr.Om, tlr.Om = self.Om.singlets_triplets()

        return slr, tlr

    def single_pole_approximation(self, i, j):
        """Return the excitation according to the
        single pole approximation. See e.g.:
        Grabo et al, Theochem 501 (2000) 353-367
        """
        for ij, kss in enumerate(self.kss):
            if kss.i == i and kss.j == j:
                return sqrt(self.Om.full[ij][ij]) * Hartree
                return self.Om.full[ij][ij] / kss.energy * Hartree

    def __str__(self):
        string = ExcitationList.__str__(self)
        string += '# derived from:\n'
        string += self.Om.kss.__str__()
        return string

    def write(self, filename=None, fh=None):
        """Write current state to a file.

        'filename' is the filename. If the filename ends in .gz,
        the file is automatically saved in compressed gzip format.

        'fh' is a filehandle. This can be used to write into already
        opened files.
        """

        if self.calculator is None:
            rank = mpi.world.rank
        else:
            rank = self.calculator.wfs.world.rank

        if rank == 0:
            if fh is None:
                f = get_filehandle(self, filename, mode='w')
            else:
                f = fh

            f.write('# ' + self.name + '\n')
            if isinstance(self.xc, str):
                xc = self.xc
            else:
                xc = self.xc.tostring()
            if xc is None:
                xc = 'RPA'
            if self.calculator is not None:
                xc += ' ' + self.calculator.get_xc_functional()
            f.write(xc + '\n')
            f.write('%g %d %g %d' % (self.kss.restrict['eps'],
                                     int(self.derivative_level),
                                     self.numscale, int(self.finegrid)) + '\n')
            self.kss.write(fh=f)
            self.Om.write(fh=f)

            if len(self):
                f.write('# Eigenvalues\n')
                f.write('{0}\n'.format(len(self)))
                for ex in self:
                    f.write(ex.outstring())
                f.write('# Eigenvectors\n')
                for ex in self:
                    for w in ex.f:
                        f.write('%g ' % w)
                    f.write('\n')

            if fh is None:
                f.close()
        mpi.world.barrier()

    def overlap(self, ov_nn, other):
        """Matrix element overlap determined from pair density overlaps.

        Parameters
        ----------
        ov_nn: array
            Wave function overlap factors from a displaced calculator.

            Index 0 corresponds to our own wavefunctions and
            index 1 to the others wavefunctions

        Returns
        -------
        ov_pp: array
            Overlap
        """
        # XXX ov_pp = self.kss.overlap(ov_nn, other.kss)
        ov_pp = self.Om.kss.overlap(ov_nn, other.Om.kss)
        self.diagonalize()
        other.diagonalize()
        # ov[pLm, pLo] = Om[pLm, :pKm]* ov[:pKm, pLo]
        return np.dot(self.Om.eigenvectors.conj(),
                      # ov[pKm, pLo] = ov[pKm, :pKo] Om[pLo, :pKo].T
                      np.dot(ov_pp, other.Om.eigenvectors.T))

    def __getitem__(self, i):
        if not self.diagonalized:
            self.diagonalize()
        return list.__getitem__(self, i)

    def __iter__(self):
        if not self.diagonalized:
            self.diagonalize()
        return list.__iter__(self)

    def __len__(self):
        if not self.diagonalized:
            self.diagonalize()
        return list.__len__(self)

    def __del__(self):
        try:
            self.timer.write(self.log.fd)
        except AttributeError:
            pass


class LrTDDFTExcitation(Excitation):

    def __init__(self, Om=None, i=None,
                 e=None, m=None, string=None):

        # multiplicity comes from Kohn-Sham contributions
        self.fij = 1

        if string is not None:
            self.fromstring(string)
            return None

        # define from the diagonalized Omega matrix
        if Om is not None:
            if i is None:
                raise RuntimeError

            ev = Om.eigenvalues[i]
            if ev < 0:
                # we reached an instability, mark it with a negative value
                self.energy = -sqrt(-ev)
            else:
                self.energy = sqrt(ev)
            self.f = Om.eigenvectors[i]
            self.kss = Om.kss

            self.kss.set_arrays()
            self.me = np.dot(self.f, self.kss.me)
            erat_k = np.sqrt(self.kss.energies / self.energy)
            wght_k = np.sqrt(self.kss.fij) * self.f
            ew_k = erat_k * wght_k
            self.mur = np.dot(ew_k, self.kss.mur)
            if self.kss.muv is not None:
                self.muv = np.dot(ew_k, self.kss.muv)
            else:
                self.muv = None
            if self.kss.magn is not None:
                self.magn = np.dot(wght_k / erat_k, self.kss.magn)
            else:
                self.magn = None

            return

        # define from energy and matrix element
        if e is not None:
            self.energy = e
            if m is None:
                raise RuntimeError
            self.me = m
            return

        raise RuntimeError

    def density_change(self, paw):
        """get the density change associated with this transition"""
        raise NotImplementedError

    def fromstring(self, string):
        l = string.split()
        self.energy = float(l.pop(0))
        if len(l) == 3:  # old writing style
            self.me = np.array([float(l.pop(0)) for i in range(3)])
        else:
            self.mur = np.array([float(l.pop(0)) for i in range(3)])
            self.me = - self.mur * sqrt(self.energy)
            self.muv = np.array([float(l.pop(0)) for i in range(3)])
            self.magn = np.array([float(l.pop(0)) for i in range(3)])

    def outstring(self):
        str = '%g ' % self.energy
        str += '  '
        for m in self.mur:
            str += '%12.4e' % m
        str += '  '
        for m in self.muv:
            str += '%12.4e' % m
        str += '  '
        for m in self.magn:
            str += '%12.4e' % m
        str += '\n'
        return str

    def __str__(self):
        m2 = np.sum(self.me * self.me)
        m = sqrt(m2)
        if m > 0:
            me = self.me / m
        else:
            me = self.me
        str = '<LrTDDFTExcitation> om=%g[eV] |me|=%g (%.2f,%.2f,%.2f)' % \
              (self.energy * Hartree, m, me[0], me[1], me[2])
        return str

    def analyse(self, min=.1):
        """Return an analysis string of the excitation"""
        osc = self.get_oscillator_strength()
        s = ('E=%.3f' % (self.energy * Hartree) + ' eV, ' +
             'f=%.5g' % osc[0] + ', (%.5g,%.5g,%.5g) ' %
             (osc[1], osc[2], osc[3]) + '\n')
        # 'R=%.5g' % self.get_rotatory_strength() + ' cgs\n')

        def sqr(x):
            return x * x
        spin = ['u', 'd']
        min2 = sqr(min)
        rest = np.sum(self.f**2)
        for f, k in zip(self.f, self.kss):
            f2 = sqr(f)
            if f2 > min2:
                s += '  %d->%d ' % (k.i, k.j) + spin[k.pspin] + ' '
                s += '%.3g \n' % f2
                rest -= f2
        s += '  rest=%.3g' % rest
        return s


def photoabsorption_spectrum(excitation_list, spectrum_file=None,
                             e_min=None, e_max=None, delta_e=None,
                             energyunit='eV',
                             folding='Gauss', width=0.1, comment=None):
    """Uniform absorption spectrum interface

    Parameters:
    ================= ===================================================
    ``exlist``        ExcitationList
    ``spectrum_file`` File name for the output file, STDOUT if not given
    ``e_min``         min. energy, set to cover all energies if not given
    ``e_max``         max. energy, set to cover all energies if not given
    ``delta_e``       energy spacing
    ``energyunit``    Energy unit, default 'eV'
    ``folding``       Gauss (default) or Lorentz
    ``width``         folding width in terms of the chosen energyunit
    ================= ===================================================
    all energies in [eV]
    """

    spectrum(exlist=excitation_list, filename=spectrum_file,
             emin=e_min, emax=e_max,
             de=delta_e, energyunit=energyunit,
             folding=folding, width=width,
             comment=comment)




# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.

import sys
import time
import math

import numpy as np
from ase.utils.timing import Timer

import gpaw.mpi as mpi


class NullTimer:
    """Compatible with Timer and StepTimer interfaces.  Does nothing."""
    def __init__(self):
        pass

    def print_info(self, calc):
        pass

    def start(self, name):
        pass

    def stop(self, name=None):
        pass

    def get_time(self, name):
        return 0.0

    def write(self, out=sys.stdout):
        pass

    def write_now(self, mark=''):
        pass

    def add(self, timer):
        pass

    def __call__(self, name):
        return self

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass


nulltimer = NullTimer()


class DebugTimer(Timer):
    def __init__(self, print_levels=1000, comm=mpi.world, txt=sys.stdout):
        Timer.__init__(self, print_levels)
        ndigits = 1 + int(math.log10(comm.size))
        self.srank = '%0*d' % (ndigits, comm.rank)
        self.txt = txt

    def start(self, name):
        Timer.start(self, name)
        abstime = time.time()
        t = self.timers[tuple(self.running)] + abstime
        self.txt.write('T%s >> %15.8f %s (%7.5fs) started\n'
                       % (self.srank, abstime, name, t))

    def stop(self, name=None):
        if name is None:
            name = self.running[-1]
        abstime = time.time()
        t = self.timers[tuple(self.running)] + abstime
        self.txt.write('T%s << %15.8f %s (%7.5fs) stopped\n'
                       % (self.srank, abstime, name, t))
        Timer.stop(self, name)


class ParallelTimer(DebugTimer):
    """Like DebugTimer but writes timings from all ranks.

    Each rank writes to timings.<rank>.txt.  Also timings.metadata.txt
    will contain information about the parallelization layout.  The idea
    is that the output from this timer can be used for plots and to
    determine bottlenecks in the parallelization.

    See the tool gpaw-plot-parallel-timings."""
    def __init__(self, prefix='timings', flush=False):
        ndigits = len(str(mpi.world.size - 1))
        ranktxt = '%0*d' % (ndigits, mpi.world.rank)
        fname = '%s.%s.txt' % (prefix, ranktxt)
        txt = open(fname, 'w', buffering=1 if flush else -1)
        DebugTimer.__init__(self, comm=mpi.world, txt=txt)
        self.prefix = prefix

    def print_info(self, calc):
        """Print information about parallelization into a file."""
        fd = open('%s.metadata.txt' % self.prefix, 'w')
        DebugTimer.print_info(self, calc)
        wfs = calc.wfs

        # We won't have to type a lot if everyone just sends all their numbers.
        myranks = np.array([wfs.world.rank, wfs.kd.comm.rank,
                            wfs.bd.comm.rank, wfs.gd.comm.rank])
        allranks = None
        if wfs.world.rank == 0:
            allranks = np.empty(wfs.world.size * 4, dtype=int)
        wfs.world.gather(myranks, 0, allranks)
        if wfs.world.rank == 0:
            for itsranks in allranks.reshape(-1, 4):
                fd.write('r=%d k=%d b=%d d=%d\n' % tuple(itsranks))
        fd.close()


class HPMTimer(Timer):
    """HPMTimer requires installation of the IBM BlueGene/P HPM
    middleware interface to the low-level UPC library. This will
    most likely only work at ANL's BlueGene/P. Must compile
    with GPAW_HPM macro in customize.py. Note that HPM_Init
    and HPM_Finalize are called in _gpaw.c and not in the Python
    interface. Timer must be called on all ranks in node, otherwise
    HPM will hang. Hence, we only call HPM_start/stop on a list
    subset of timers."""

    top_level = 'GPAW.calculator'  # HPM needs top level timer
    compatible = ['Initialization', 'SCF-cycle']

    def __init__(self):
        Timer.__init__(self)
        from _gpaw import hpm_start, hpm_stop
        self.hpm_start = hpm_start
        self.hpm_stop = hpm_stop
        hpm_start(self.top_level)

    def start(self, name):
        Timer.start(self, name)
        if name in self.compatible:
            self.hpm_start(name)

    def stop(self, name=None):
        Timer.stop(self, name)
        if name in self.compatible:
            self.hpm_stop(name)

    def write(self, out=sys.stdout):
        Timer.write(self, out)
        self.hpm_stop(self.top_level)


class CrayPAT_timer(Timer):
    """Interface to CrayPAT API. In addition to regular timers,
    the corresponding regions are profiled by CrayPAT. The gpaw-python has
    to be compiled under CrayPAT.
    """

    def __init__(self, print_levels=4):
        Timer.__init__(self, print_levels)
        from _gpaw import craypat_region_begin, craypat_region_end
        self.craypat_region_begin = craypat_region_begin
        self.craypat_region_end = craypat_region_end
        self.regions = {}
        self.region_id = 5  # leave room for regions in C

    def start(self, name):
        Timer.start(self, name)
        if name in self.regions:
            id = self.regions[name]
        else:
            id = self.region_id
            self.regions[name] = id
            self.region_id += 1
        self.craypat_region_begin(id, name)

    def stop(self, name=None):
        Timer.stop(self, name)
        id = self.regions[name]
        self.craypat_region_end(id)



import numpy as np

from gpaw.mpi import world
from gpaw.tddft.units import as_to_au, eV_to_au


def create_laser(name, **kwargs):
    if isinstance(name, Laser):
        return name
    elif isinstance(name, dict):
        kwargs.update(name)
        return create_laser(**kwargs)
    elif name == 'GaussianPulse':
        return GaussianPulse(**kwargs)
    elif name == 'SumLaser':
        return SumLaser(**kwargs)
    else:
        raise ValueError('Unknown laser: %s' % name)


class Laser(object):
    def __init__(self):
        pass

    def strength(self, time):
        return 0.0

    def derivative(self, time):
        return 0.0

    def fourier(self, omega):
        return 0.0

    def write(self, fname, time_t):
        """
        Write the values of the pulse to a file.

        Parameters
        ----------
        fname
            filename
        time_t
            times in attoseconds
        """
        if world.rank != 0:
            return
        time_t = time_t * as_to_au
        strength_t = self.strength(time_t)
        derivative_t = self.derivative(time_t)
        fmt = '%12.6f %20.10e %20.10e'
        header = '{:^10} {:^20} {:^20}'.format('time', 'strength',
                                               'derivative')
        np.savetxt(fname, np.stack((time_t, strength_t, derivative_t)).T,
                   fmt=fmt, header=header)


class SumLaser(Laser):
    def __init__(self, *lasers):
        self.laser_i = []
        dict_i = []
        for laser in lasers:
            laser = create_laser(laser)
            self.laser_i.append(laser)
            dict_i.append(laser.todict())
        self.dict = dict(name='SumLaser',
                         lasers=dict_i)

    def strength(self, time):
        s = 0.0
        for laser in self.laser_i:
            s += laser.strength(time)
        return s

    def fourier(self, omega):
        s = 0.0
        for laser in self.laser_i:
            s += laser.fourier(omega)
        return s

    def todict(self):
        return self.dict


class GaussianPulse(Laser):
    r"""
    Laser pulse with Gaussian envelope:

    .. math::

        g(t) = s_0 \sin(\omega_0 (t - t_0)) \exp(-\sigma^2 (t - t_0)^2 / 2)


    Parameters
    ----------
    strength: float
        value of :math:`s_0` in atomic units
    time0: float
        value of :math:`t_0` in attoseconds
    frequency: float
        value of :math:`\omega_0` in eV
    sigma: float
        value of :math:`\sigma` in eV
    sincos: 'sin' or 'cos'
        use sin or cos function
    stoptime: float
        pulse is set to zero after this value (in attoseconds)
    """

    def __init__(self, strength, time0, frequency, sigma, sincos='sin',
                 stoptime=np.inf):
        self.dict = dict(name='GaussianPulse',
                         strength=strength,
                         time0=time0,
                         frequency=frequency,
                         sigma=sigma,
                         sincos=sincos)
        self.s0 = strength
        self.t0 = time0 * as_to_au
        self.omega0 = frequency * eV_to_au
        self.sigma = sigma * eV_to_au
        self.stoptime = stoptime * as_to_au
        assert sincos in ['sin', 'cos']
        self.sincos = sincos

    def strength(self, t):
        """
        Return the value of the pulse :math:`g(t)`.

        Parameters
        ----------
        t
            time in atomic units

        Returns
        -------
        The value of the pulse.
        """
        s = self.s0 * np.exp(-0.5 * self.sigma**2 * (t - self.t0)**2)
        if self.sincos == 'sin':
            s *= np.sin(self.omega0 * (t - self.t0))
        else:
            s *= np.cos(self.omega0 * (t - self.t0))
        flt = t < self.stoptime

        return s * flt

    def derivative(self, t):
        """
        Return the derivative of the pulse :math:`g'(t)`.

        Parameters
        ----------
        t
            time in atomic units

        Returns
        -------
        The derivative of the pulse.
        """
        dt = t - self.t0
        s = self.s0 * np.exp(-0.5 * self.sigma**2 * dt**2)
        if self.sincos == 'sin':
            s *= (-self.sigma**2 * dt * np.sin(self.omega0 * dt) +
                  self.omega0 * np.cos(self.omega0 * dt))
        else:
            s *= (-self.sigma**2 * dt * np.cos(self.omega0 * dt) +
                  -self.omega0 * np.sin(self.omega0 * dt))
        return s

    def fourier(self, omega):
        r"""
        Return Fourier transform of the pulse :math:`g(\omega)`.

        Parameters
        ----------
        omega
            frequency in atomic units

        Returns
        -------
        Fourier transform of the pulse.
        """
        s = (self.s0 * np.sqrt(np.pi / 2) / self.sigma *
             np.exp(-0.5 * (omega - self.omega0)**2 / self.sigma**2) *
             np.exp(1.0j * self.t0 * omega))
        if self.sincos == 'sin':
            s *= 1.0j
        return s

    def todict(self):
        return self.dict



"""Scissors operator for LCAO."""
from typing import Sequence, Tuple

import numpy as np
from ase.units import Ha

from gpaw.hamiltonian import Hamiltonian
from gpaw.typing import Array2D, Array3D
from gpaw.kpoint import KPoint
from gpaw.wavefunctions.base import WaveFunctions

from .eigensolver import DirectLCAO


class Scissors(DirectLCAO):
    def __init__(self, shifts: Sequence[Tuple[float, float, int]]):
        """Scissors-operator eigensolver.

        The *shifts* are given as a sequence of tuples::

            [(<shift for occupied states>,
              <shift for unoccupied states>,
              <number of atoms>),
             ...]

        Here we open a gap for states on atoms with indices 3, 4 and 5:

        >>> eigensolver = Scissors([(0.0, 0.0, 3),
        ...                         (-0.5, 0.5, 3)])

        """
        DirectLCAO.__init__(self)
        self.shifts = []
        for homo, lumo, natoms in shifts:
            self.shifts.append((homo / Ha, lumo / Ha, natoms))

    def write(self, writer):
        writer.write(name='lcao')

    def __repr__(self):
        txt = DirectLCAO.__repr__(self)
        txt += '\n    Scissors operators:\n'
        a1 = 0
        for homo, lumo, natoms in self.shifts:
            a2 = a1 + natoms
            txt += (f'      Atoms {a1}-{a2 - 1}: '
                    f'VB: {homo * Ha:+.3f} eV, '
                    f'CB: {lumo * Ha:+.3f} eV\n')
            a1 = a2
        return txt

    def calculate_hamiltonian_matrix(self,
                                     ham: Hamiltonian,
                                     wfs: WaveFunctions,
                                     kpt: KPoint,
                                     Vt_xMM: Array3D = None,
                                     root: int = -1,
                                     add_kinetic: bool = True) -> Array2D:
        """Add scissors operator."""
        H_MM = DirectLCAO.calculate_hamiltonian_matrix(
            self, ham, wfs, kpt, Vt_xMM, root, add_kinetic)
        if kpt.C_nM is None:
            return H_MM

        S_MM = wfs.S_qMM[kpt.q]
        assert abs(S_MM - S_MM.T.conj()).max() < 1e-10

        nocc = ham.setups.nvalence // 2
        C_nM = kpt.C_nM
        iC_Mn = np.linalg.inv(C_nM)
        Co_nM = C_nM.copy()
        Co_nM[nocc:] = 0.0
        Cu_nM = C_nM.copy()
        Cu_nM[:nocc] = 0.0

        M1 = 0
        a1 = 0
        for homo, lumo, natoms in self.shifts:
            a2 = a1 + natoms
            M2 = M1 + sum(setup.nao for setup in ham.setups[a1:a2])

            D_MM = np.zeros_like(S_MM)
            D_MM[M1:M2, M1:M2] = S_MM[M1:M2, M1:M2]

            H_MM += iC_Mn @ (
                Co_nM @ D_MM @ Co_nM.T.conj() * homo +
                Cu_nM @ D_MM @ Cu_nM.T.conj() * lumo) @ iC_Mn.T.conj()

            a1 = a2
            M1 = M2

        return H_MM



from __future__ import annotations

# General modules
from abc import abstractmethod

import numpy as np

# GPAW modules
from gpaw.sphere.integrate import (integrate_lebedev,
                                   radial_truncation_function,
                                   spherical_truncation_function_collection,
                                   default_spherical_drcut,
                                   find_volume_conserving_lambd)

from gpaw.response import ResponseGroundStateAdapter, ResponseContext
from gpaw.response.frequencies import ComplexFrequencyDescriptor
from gpaw.response.chiks import ChiKSCalculator, smat
from gpaw.response.localft import (LocalFTCalculator,
                                   add_LSDA_Wxc,
                                   add_spin_polarization,
                                   add_LSDA_spin_splitting,
                                   extract_micro_setup)
from gpaw.response.site_kernels import SiteKernels
from gpaw.response.pair_functions import SingleQPWDescriptor, PairFunction
from gpaw.response.pair_integrator import PairFunctionIntegrator
from gpaw.response.matrix_elements import (SiteMatrixElementCalculator,
                                           SitePairDensityCalculator,
                                           SitePairSpinSplittingCalculator)

# ASE modules
from ase.units import Hartree, Bohr

from ase.neighborlist import natural_cutoffs, build_neighbor_list


class IsotropicExchangeCalculator:
    r"""Calculator class for the Heisenberg exchange constants

    _           2
    J^ab(q) = - ‾‾ B^(xc†) K^(a†)(q) χ_KS^('+-)(q) K^b(q) B^(xc)            (1)
                V0

    calculated for an isotropic system in a plane wave representation using
    the magnetic force theorem within second order perturbation theory, see
    [J. Phys.: Condens. Matter 35 (2023) 105802].

    Entering the formula for the isotropic exchange constant at wave vector q
    between sublattice a and b is the unit cell volume V0, the functional
    derivative of the (LDA) exchange-correlation energy with respect to the
    magnitude of the magnetization B^(xc), the sublattice site kernels K^a(q)
    and K^b(q) as well as the reactive part of the static transverse magnetic
    susceptibility of the Kohn-Sham system χ_KS^('+-)(q).

    NB: To achieve numerical stability of the plane-wave implementation, we
    use instead the following expression to calculate exchange parameters:

    ˷           2
    J^ab(q) = - ‾‾ W_xc^(z†) K^(a†)(q) χ_KS^('+-)(q) K^b(q) W_xc^z          (2)
                V0

    We do this since B^(xc)(r) = -|W_xc^z(r)| is nonanalytic in points of space
    where the spin-polarization changes sign, why it is problematic to evaluate
    Eq. (1) numerically within a plane-wave representation.
    If the site partitionings only include spin-polarization of the same sign,
    Eqs. (1) and (2) should yield identical exchange parameters, but for
    antiferromagnetically aligned sites, the coupling constants differ by a
    sign.

    The site kernels encode the partitioning of real space into sites of the
    Heisenberg model. This is not a uniquely defined procedure, why the user
    has to define them externally through the SiteKernels interface."""

    def __init__(self,
                 chiks_calc: ChiKSCalculator,
                 localft_calc: LocalFTCalculator):
        """Construct the IsotropicExchangeCalculator object."""
        # Check that chiks has the assumed properties
        assumed_props = dict(
            gammacentered=True,
            nblocks=1
        )
        for key, item in assumed_props.items():
            assert getattr(chiks_calc, key) == item, \
                f'Expected chiks.{key} == {item}. '\
                f'Got: {getattr(chiks_calc, key)}'

        self.chiks_calc = chiks_calc
        self.context = chiks_calc.context

        # Check assumed properties of the LocalFTCalculator
        assert localft_calc.context is self.context
        assert localft_calc.gs is chiks_calc.gs
        self.localft_calc = localft_calc

        # W_xc^z buffer
        self._Wxc_G = None

        # χ_KS^('+-) buffer
        self._chiksr = None

    def __call__(self, q_c, site_kernels: SiteKernels, txt=None):
        """Calculate the isotropic exchange constants for a given wavevector.

        Parameters
        ----------
        q_c : nd.array
            Wave vector q in relative coordinates
        site_kernels : SiteKernels
            Site kernels instance defining the magnetic sites of the crystal
        txt : str
            Separate file to store the chiks calculation output in (optional).
            If not supplied, the output will be written to the standard text
            output location specified when initializing chiks.

        Returns
        -------
        J_abp : nd.array (dtype=complex)
            Isotropic Heisenberg exchange constants between magnetic sites a
            and b for all the site partitions p given by the site_kernels.
        """
        # Get ingredients
        Wxc_G = self.get_Wxc()
        chiksr = self.get_chiksr(q_c, txt=txt)
        qpd, chiksr_GG = chiksr.qpd, chiksr.array[0]  # array = chiksr_zGG
        V0 = qpd.gd.volume

        # Allocate an array for the exchange constants
        nsites = site_kernels.nsites
        J_pab = np.empty(site_kernels.shape + (nsites,), dtype=complex)

        # Compute exchange coupling
        for J_ab, K_aGG in zip(J_pab, site_kernels.calculate(qpd)):
            for a in range(nsites):
                for b in range(nsites):
                    J = np.conj(Wxc_G) @ np.conj(K_aGG[a]).T @ chiksr_GG \
                        @ K_aGG[b] @ Wxc_G
                    J_ab[a, b] = - 2. * J / V0

        # Transpose to have the partitions index last
        J_abp = np.transpose(J_pab, (1, 2, 0))

        return J_abp * Hartree  # Convert from Hartree to eV

    def get_Wxc(self):
        """Get B^(xc)_G from buffer."""
        if self._Wxc_G is None:  # Calculate if buffer is empty
            self._Wxc_G = self._calculate_Wxc()

        return self._Wxc_G

    def _calculate_Wxc(self):
        """Calculate the Fourier transform W_xc^z(G)."""
        # Create a plane wave descriptor encoding the plane wave basis. Input
        # q_c is arbitrary, since we are assuming that chiks.gammacentered == 1
        qpd0 = self.chiks_calc.get_pw_descriptor([0., 0., 0.])

        return self.localft_calc(qpd0, add_LSDA_Wxc)

    def get_chiksr(self, q_c, txt=None):
        """Get χ_KS^('+-)(q) from buffer."""
        q_c = np.asarray(q_c)

        # Calculate if buffer is empty or a new q-point is given
        if self._chiksr is None or not np.allclose(q_c, self._chiksr.q_c):
            self._chiksr = self._calculate_chiksr(q_c, txt=txt)

        return self._chiksr

    def _calculate_chiksr(self, q_c, txt=None):
        r"""Use the ChiKSCalculator to calculate the reactive part of the
        static Kohn-Sham susceptibility χ_KS^('+-)(q).

        First, the dynamic Kohn-Sham susceptibility

                                 __  __
                              1  \   \        f_nk↑ - f_mk+q↓
        χ_KS,GG'^+-(q,ω+iη) = ‾  /   /  ‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
                              V  ‾‾  ‾‾ ħω - (ε_mk+q↓ - ε_nk↑) + iħη
                                 k  n,m
                                        x n_nk↑,mk+q↓(G+q) n_mk+q↓,nk↑(-G'-q)

        is calculated in the static limit ω=0 and without broadening η=0. Then,
        the reactive part (see [PRB 103, 245110 (2021)]) is extracted:

                              1
        χ_KS,GG'^(+-')(q,z) = ‾ [χ_KS,GG'^+-(q,z) + χ_KS,-G'-G^-+(-q,-z*)].
                              2
        """
        # Initiate new output file, if supplied
        if txt is not None:
            self.context.new_txt_and_timer(txt)

        zd = ComplexFrequencyDescriptor.from_array([0. + 0.j])
        chiks = self.chiks_calc.calculate('+-', q_c, zd)
        if np.allclose(q_c, 0.):
            chiks.symmetrize_reciprocity()

        # Take the reactive part
        chiksr = chiks.copy_reactive_part()

        return chiksr


class AtomicSiteData:
    r"""Data object for spherical atomic sites."""

    def __init__(self, gs: ResponseGroundStateAdapter,
                 indices, radii):
        """Construct the atomic site data object from a ground state adapter.

        Parameters
        ----------
        indices : 1D array-like
            Atomic index A for each site index a.
        radii : 2D array-like
            Atomic radius rc for each site index a and partitioning p.
        """
        self.A_a = np.asarray(indices)
        assert self.A_a.ndim == 1
        assert len(np.unique(self.A_a)) == len(self.A_a)

        # Parse the input atomic radii
        rc_ap = np.asarray(radii)
        assert rc_ap.ndim == 2
        assert rc_ap.shape[0] == len(self.A_a)
        # Convert radii to internal units (Å to Bohr)
        self.rc_ap = rc_ap / Bohr

        self.nsites = len(self.A_a)
        self.npartitions = self.rc_ap.shape[1]
        self.shape = (self.nsites, self.npartitions)

        assert self._in_valid_site_radii_range(gs), \
            'Please provide site radii in the valid range, see '\
            'AtomicSiteData.valid_site_radii_range()'

        # Extract the scaled positions and microsetups for each atomic site
        self.spos_ac = gs.spos_ac[self.A_a]
        self.microsetup_a = [extract_micro_setup(gs, A) for A in self.A_a]

        # Extract pseudo density on the fine real-space grid
        self.finegd = gs.finegd
        self.nt_sr = gs.nt_sr

        # Set up the atomic truncation functions which define the sites based
        # on the coarse real-space grid
        self.gd = gs.gd
        self.drcut = default_spherical_drcut(self.gd)
        self.lambd_ap = np.array(
            [[find_volume_conserving_lambd(rcut, self.drcut)
              for rcut in rc_p] for rc_p in self.rc_ap])
        self.stfc = spherical_truncation_function_collection(
            self.finegd, self.spos_ac, self.rc_ap, self.drcut, self.lambd_ap)

    @staticmethod
    def _valid_site_radii_range(gs):
        """For each atom in gs, determine the valid site radii range in Bohr.

        The lower bound is determined by the spherical truncation width, when
        truncating integrals on the real-space grid.
        The upper bound is determined by the distance to the nearest
        augmentation sphere.
        """
        atoms = gs.atoms
        drcut = default_spherical_drcut(gs.gd)
        rmin_A = np.array([drcut / 2] * len(atoms))

        # Find neighbours based on covalent radii
        cutoffs = natural_cutoffs(atoms, mult=2)
        neighbourlist = build_neighbor_list(atoms, cutoffs,
                                            self_interaction=False)
        # Determine rmax for each atom
        augr_A = gs.get_aug_radii()
        rmax_A = []
        for A in range(len(atoms)):
            pos = atoms.positions[A]
            # Calculate the distance to the augmentation sphere of each
            # neighbour
            aug_distances = []
            for An, offset in zip(*neighbourlist.get_neighbors(A)):
                posn = atoms.positions[An] + offset @ atoms.get_cell()
                dist = np.linalg.norm(posn - pos) / Bohr  # Å -> Bohr
                aug_dist = dist - augr_A[An]
                assert aug_dist > 0.
                aug_distances.append(aug_dist)
            # In order for PAW corrections to be valid, we need a sphere of
            # radius rcut not to overlap with any neighbouring augmentation
            # spheres
            rmax_A.append(min(aug_distances))
        rmax_A = np.array(rmax_A)

        return rmin_A, rmax_A

    @staticmethod
    def valid_site_radii_range(gs):
        """Get the valid site radii for all atoms in a given ground state."""
        rmin_A, rmax_A = AtomicSiteData._valid_site_radii_range(gs)
        # Convert to external units (Bohr to Å)
        return rmin_A * Bohr, rmax_A * Bohr

    def _in_valid_site_radii_range(self, gs):
        rmin_A, rmax_A = AtomicSiteData._valid_site_radii_range(gs)
        for a, A in enumerate(self.A_a):
            if not np.all(
                    np.logical_and(
                        self.rc_ap[a] > rmin_A[A] - 1e-8,
                        self.rc_ap[a] < rmax_A[A] + 1e-8)):
                return False
        return True

    def calculate_magnetic_moments(self):
        """Calculate the magnetic moments at each atomic site."""
        magmom_ap = self.integrate_local_function(add_spin_polarization)
        return magmom_ap

    def calculate_spin_splitting(self):
        r"""Calculate the spin splitting Δ^(xc) for each atomic site."""
        dxc_ap = self.integrate_local_function(add_LSDA_spin_splitting)
        return dxc_ap * Hartree  # return the splitting in eV

    def integrate_local_function(self, add_f):
        r"""Integrate a local function f[n](r) = f(n(r)) over the atomic sites.

        For every site index a and partitioning p, the integral is defined via
        a smooth truncation function θ(|r-r_a|<rc_ap):

               /
        f_ap = | dr θ(|r-r_a|<rc_ap) f(n(r))
               /
        """
        out_ap = np.zeros(self.shape, dtype=float)
        self._integrate_pseudo_contribution(add_f, out_ap)
        self._integrate_paw_correction(add_f, out_ap)
        return out_ap

    def _integrate_pseudo_contribution(self, add_f, out_ap):
        """Calculate the pseudo contribution to the atomic site integrals.

        For local functions of the density, the pseudo contribution is
        evaluated by a numerical integration on the real-space grid:

        ̰       /
        f_ap = | dr θ(|r-r_a|<rc_ap) f(ñ(r))
               /
        """
        # Evaluate the local function on the real-space grid
        ft_r = self.finegd.zeros()
        add_f(self.finegd, self.nt_sr, ft_r)

        # Integrate θ(|r-r_a|<rc_ap) f(ñ(r))
        ftdict_ap = {a: np.empty(self.npartitions) for a in range(self.nsites)}
        self.stfc.integrate(ft_r, ftdict_ap)

        # Add pseudo contribution to the output array
        for a in range(self.nsites):
            out_ap[a] += ftdict_ap[a]

    def _integrate_paw_correction(self, add_f, out_ap):
        """Calculate the PAW correction to an atomic site integral.

        The PAW correction is evaluated on the atom centered radial grid, using
        the all-electron and pseudo densities generated from the partial waves:

                /
        Δf_ap = | r^2 dr θ(r<rc_ap) [f(n_a(r)) - f(ñ_a(r))]
                /
        """
        for a, (microsetup, rc_p, lambd_p) in enumerate(zip(
                self.microsetup_a, self.rc_ap, self.lambd_ap)):
            # Evaluate the PAW correction and integrate angular components
            df_ng = microsetup.evaluate_paw_correction(add_f)
            df_g = integrate_lebedev(df_ng)
            for p, (rcut, lambd) in enumerate(zip(rc_p, lambd_p)):
                # Evaluate the smooth truncation function
                theta_g = radial_truncation_function(
                    microsetup.rgd.r_g, rcut, self.drcut, lambd)
                # Integrate θ(r) Δf(r) on the radial grid
                out_ap[a, p] += microsetup.rgd.integrate_trapz(df_g * theta_g)


class StaticSitePairFunction(PairFunction):
    """Data object for static site pair functions."""

    def __init__(self,
                 qpd: SingleQPWDescriptor,
                 atomic_site_data: AtomicSiteData):
        self.qpd = qpd
        self.q_c = qpd.q_c

        self.atomic_site_data = atomic_site_data

        self.array = self.zeros()

    @property
    def shape(self):
        nsites = self.atomic_site_data.nsites
        npartitions = self.atomic_site_data.npartitions
        return nsites, nsites, npartitions
        
    def zeros(self):
        return np.zeros(self.shape, dtype=complex)


class TwoParticleSiteSumRuleCalculator(PairFunctionIntegrator):
    r"""Calculator for two-particle site sum rules.

    For any set of site matrix elements f^a and g^b, one may define a two-
    particle site sum rule based on the lattice Fourier transformed quantity:
                     __  __   __
                 1   \   \    \   /
    ̄x_ab^z(q) = ‾‾‾  /   /    /   | σ^j_ss' (f_nks - f_n'k+qs')
                N_k  ‾‾  ‾‾   ‾‾  \                                       \
                     k  n,n' s,s'   × f^a_(nks,n'k+qs') g^b_(n'k+qs',nks) |
                                                                          /
    where σ^j is a Pauli matrix with j∊{0,+,-,z}.
    """

    def __init__(self,
                 gs: ResponseGroundStateAdapter,
                 context: ResponseContext | None = None,
                 nblocks: int = 1,
                 nbands: int | None = None):
        """Construct the two-particle site sum rule calculator."""
        if context is None:
            context = ResponseContext()
        super().__init__(gs, context,
                         nblocks=nblocks,
                         # Disable use of symmetries for now. The sum rule site
                         # magnetization symmetries can always be derived and
                         # implemented at a later stage.
                         disable_point_group=True,
                         disable_time_reversal=True)

        self.nbands = nbands
        self.matrix_element_calc1: SiteMatrixElementCalculator | None = None
        self.matrix_element_calc2: SiteMatrixElementCalculator | None = None

    def __call__(self, q_c, atomic_site_data: AtomicSiteData):
        """Calculate the site sum rule for a given wave vector q_c."""
        # Set up calculators for the f^a and g^b matrix elements
        mecalc1, mecalc2 = self.create_matrix_element_calculators(
            atomic_site_data)
        self.matrix_element_calc1 = mecalc1
        self.matrix_element_calc2 = mecalc2

        spincomponent = self.get_spincomponent()
        transitions = self.get_band_and_spin_transitions(
            spincomponent, nbands=self.nbands, bandsummation='double')
        self.context.print(self.get_info_string(
            q_c, self.nbands, len(transitions)))

        # Set up data object (without a plane-wave representation, which is
        # irrelevant in this case)
        qpd = self.get_pw_descriptor(q_c, ecut=1e-3)
        site_pair_function = StaticSitePairFunction(qpd, atomic_site_data)

        # Perform actual calculation
        self._integrate(site_pair_function, transitions)

        return site_pair_function.array

    @abstractmethod
    def create_matrix_element_calculators(self, atomic_site_data):
        """Create the desired site matrix element calculators."""

    @abstractmethod
    def get_spincomponent(self):
        """Define how to rotate the spins via the spin component (μν)."""

    def add_integrand(self, kptpair, weight, site_pair_function):
        r"""Add the site sum rule integrand of the outer k-point integral.

        With
                       __
                    1  \
        ̄x_ab^z(q) = ‾  /  (...)_k
                    V  ‾‾
                       k

        the integrand is given by
                     __   __
                     \    \   /
        (...)_k = V0 /    /   | σ^j_ss' (f_nks - f_n'k+qs')
                     ‾‾   ‾‾  \                                       \
                    n,n' s,s'   × f^a_(nks,n'k+qs') g^b_(n'k+qs',nks) |
                                                                      /

        where V0 is the cell volume.
        """
        # Calculate site matrix elements
        qpd = site_pair_function.qpd
        matrix_element1 = self.matrix_element_calc1(kptpair, qpd)
        if self.matrix_element_calc2 is self.matrix_element_calc1:
            matrix_element2 = matrix_element1
        else:
            matrix_element2 = self.matrix_element_calc2(kptpair, qpd)

        # Calculate the product between the Pauli matrix and the occupational
        # differences
        sigma = self.get_pauli_matrix()
        s1_myt, s2_myt = kptpair.get_local_spin_indices()
        sigma_myt = sigma[s1_myt, s2_myt]
        df_myt = kptpair.ikpt1.f_myt - kptpair.ikpt2.f_myt
        sigmadf_myt = sigma_myt * df_myt

        # Calculate integrand
        f_mytap = matrix_element1.local_array_view
        g_mytap = matrix_element2.local_array_view
        fgcc_mytabp = f_mytap[:, :, np.newaxis] * g_mytap.conj()[:, np.newaxis]
        # Sum over local transitions
        integrand_abp = np.einsum('t, tabp -> abp', sigmadf_myt, fgcc_mytabp)
        # Sum over distributed transitions
        kptpair.tblocks.blockcomm.sum(integrand_abp)

        # Add integrand to output array
        site_pair_function.array[:] += self.gs.volume * weight * integrand_abp

    @abstractmethod
    def get_pauli_matrix(self):
        """Get the desired Pauli matrix σ^j_ss'."""

    def get_info_string(self, q_c, nbands, nt):
        """Get information about the calculation"""
        info_list = ['',
                     'Calculating two-particle site sum rule with:'
                     f'    q_c: [{q_c[0]}, {q_c[1]}, {q_c[2]}]',
                     self.get_band_and_transitions_info_string(nbands, nt),
                     '',
                     self.get_basic_info_string()]
        return '\n'.join(info_list)


class TwoParticleSiteMagnetizationCalculator(TwoParticleSiteSumRuleCalculator):
    r"""Calculator for the two-particle site magnetization sum rule.

    The site magnetization can be calculated from the site pair densities via
    the following sum rule [publication in preparation]:
                     __  __
                 1   \   \
    ̄n_ab^z(q) = ‾‾‾  /   /  (f_nk↑ - f_mk+q↓) n^a_(nk↑,mk+q↓) n^b_(mk+q↓,nk↑)
                N_k  ‾‾  ‾‾
                     k   n,m

              = δ_(a,b) n_a^z

    This is directly related to the sum rule of the χ^(+-) spin component of
    the four-component susceptibility tensor.
    """
    def create_matrix_element_calculators(self, atomic_site_data):
        site_pair_density_calc = SitePairDensityCalculator(
            self.gs, self.context, atomic_site_data)
        return site_pair_density_calc, site_pair_density_calc

    def get_spincomponent(self):
        return '+-'

    def get_pauli_matrix(self):
        return smat('+')


class TwoParticleSiteSpinSplittingCalculator(
        TwoParticleSiteMagnetizationCalculator):
    r"""Calculator for the two-particle site spin splitting sum rule.

    The site spin splitting can be calculated from the site pair density and
    site pair spin splitting via the following sum rule [publication in
    preparation]:
                          __  __
    ˍ                 1   \   \  /
    Δ^(xc)_ab^z(q) = ‾‾‾  /   /  | (f_nk↑ - f_mk+q↓)
                     N_k  ‾‾  ‾‾ \                                        \
                          k   n,m  × Δ^(xc,a)_(nk↑,mk+q↓) n^b_(mk+q↓,nk↑) |
                                                                          /
              = δ_(a,b) Δ^(xc)_a^z
    """
    def create_matrix_element_calculators(self, atomic_site_data):
        site_pair_spin_splitting_calc = SitePairSpinSplittingCalculator(
            self.gs, self.context, atomic_site_data)
        site_pair_density_calc = SitePairDensityCalculator(
            self.gs, self.context, atomic_site_data)
        return site_pair_spin_splitting_calc, site_pair_density_calc

    def __call__(self, *args):
        dxc_abp = super().__call__(*args)
        return dxc_abp * Hartree  # Ha -> eV



import pytest
from ase import Atoms
from ase.io import read

from gpaw import GPAW, FermiDirac


@pytest.mark.ci
def test_no_cell():
    with pytest.raises(ValueError):
        H = Atoms('H', calculator=GPAW(mode='fd'))
        H.get_potential_energy()


@pytest.mark.later
def test_read_txt(in_tmp_dir):
    a = 2.0
    calc = GPAW(
        mode='fd',
        gpts=(12, 12, 12),
        txt='H.txt',
        occupations=FermiDirac(0.0))
    H = Atoms('H',
              cell=(a, a, a),
              pbc=True,
              calculator=calc)
    e0 = H.get_potential_energy()

    H = read('H.txt')
    assert H.get_potential_energy() == pytest.approx(e0)

    energy_tolerance = 0.001
    assert e0 == pytest.approx(-6.5577, abs=energy_tolerance)
